# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
from abc import abstractmethod


class ResultInfo(object):
    """
    Object holding the result of the discovery on particular dimension.
    The purpose of the object is to summarize and return the discovery results

    :param object: _description_
    :type object: _type_
    """ 
    def __init__(self):
        """_summary_
        """        
    @abstractmethod
    def getDiscoveryResult(self) -> str:
        """
        Return the discovery results summarized as a string

        :return: Discovery result
        :rtype: str
        """        
        return NotImplementedError


# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
from __future__ import annotations

from flasgger import Swagger
from flask import Flask

from .exceptions.badrequest import BadRequestException
from .exceptions.notfound import NotFoundException
from .exceptions.validation import ValidationException
from .routes.sax_routes import sax_routes
from .state.state import State


def create_app():
    # Create and configure app 
    app = Flask(__name__)    
    app.state = State()

    # Register Routes    
    app.register_blueprint(sax_routes)
    
    @app.errorhandler(BadRequestException)
    def handle_bad_request(err):
        return {"message": str(err)}, 400

    @app.errorhandler(ValidationException)
    def handle_validation_exception(err):
        return {"message": str(err)}, 422

    @app.errorhandler(NotFoundException)
    def handle_not_found_exception(err):
        return {"message": str(err)}, 404
    
    Swagger(app)

    return app

 
 


# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
from typing import Optional

import pandas as pd

from sax.core.utils.constants import Constants
from .base_anchor import BaseAnchor
from ...causal_process_discovery.algorithms.base_causal_alg import CausalResultInfo
from ...causal_process_discovery.algorithms.lingam import LingamImpl
from ...causal_process_discovery.causal_constants import DEFAULT_VARIANT, Algorithm
from ...causal_process_discovery.prior_knowledge import PriorKnowledge
from ...process_data.raw_event_data import RawEventData


class ChainAnchorTransformer(BaseAnchor):
    '''
    Causal chain modality transformer - transforms process data object from event log format to a tabular format which represents each activity
    as a column in the data frame, where the value of the column is the accumulated duration of this activity from trace start time.
    The chain modality is described here <>
    '''    

    def apply(self, dataObject: RawEventData,variant: Optional[Algorithm] = DEFAULT_VARIANT,prior_knowledge: Optional[bool]=False) -> CausalResultInfo:
        """
        Transform raw event log in tabular form (the provided event log) to row format, so that each trace is represented by a row and each activity column value holds the accimulated duration
        from trace start time. After the transformation, applies the chosen causal discovery algorithm variant to discover causal execution dependencies
        among chained durations.

        :param dataObject: event log
        :type dataObject: RawEventData
        :param variant: algorithm variant to apply, defaults to LINGAM
        :type variant: Optional[Algorithm], optional
        :param prior_knowledge: whether to apply prior knowledge, defaults to False
        :type prior_knowledge: Optional[bool], optional
        :raises TypeError: _description_
        :return: _description_
        :rtype: CausalResultInfo
        """
        if type(dataObject) not in [RawEventData]: raise TypeError("the method can be applied only to an object of type RawEventData!")
        
        dataObject = dataObject.transposeToTabular()      
        
          
        #get the main properties data - process id and activities columns from the dataframe
        df = dataObject.getCaseAndActivitiesData() #TODO can change only to getActivitiesData() and update the code      
        start_time_column_name = dataObject.getMandatoryProperties()[Constants.STARTTIME_COLUMN] 
        caseIdColumnName = dataObject.getCaseIdColumnName()              

        # Drop the 'Id' column to create activities_df       
        activities_df = df.drop(caseIdColumnName, axis=1).copy()          
        activities_df = activities_df.apply(pd.to_datetime, utc=True)     
        start_times = activities_df[start_time_column_name].copy()     
        activities_df= activities_df.drop(start_time_column_name,axis=1)
                         
        first_activity_time = start_times                  
        
        # Create a new DataFrame for storing time differences
        time_difference_df = activities_df.copy()
        for col in activities_df.columns:
             time_difference_df[col] = (activities_df[col] - first_activity_time).dt.total_seconds()*1000
                     
       
        #create and run the algorithm        
        args = {"data": time_difference_df}
        if prior_knowledge:
            args["prior_knowledge"] = PriorKnowledge(time_difference_df)
        
        if variant == Algorithm.LINGAM:
            algorithm = LingamImpl(**args)
        result  = algorithm.run()       

        return result  


# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
class LLMTimeoutException(Exception):
    """Exception risen when there is timeout problem when invoking external LLM
    """ 
    def __init__(self, message, details):
        self.message = message
        self.details = details
        super().__init__(self.message)


# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
import os
from genai import Client, Credentials
from genai.extensions.langchain import LangChainInterface
from genai.extensions.langchain import LangChainEmbeddingsInterface
from genai.schema import TextEmbeddingParameters
from sax.core.synthesis.llms.base_llm import BaseLLM

class WatsonXLLM(BaseLLM):
    """WatsonX LLM wrapper, implementation of BaseLLM
    """
    def __init__(self, model, temperature): 
        self.genai_key = os.getenv("GENAI_KEY")                
        self.genai_api= os.getenv("GENAI_API")       
        if not self.genai_key:
            raise ValueError("The API key for GenAI is not set. Please set the GENAI_KEY environment variable.")
        if not self.genai_api:
            raise ValueError("The API URL for GenAI is not set. Please set the GENAI_API environment variable.")
        super().__init__(model,temperature) 
        
        credentials = Credentials(api_key=self.genai_key, api_endpoint=self.genai_api)
        client = Client(credentials=credentials)
        client = Client(credentials=Credentials.from_env())
        self._model = LangChainInterface(model_id=model, client=client)                
        self._embeddingModel =LangChainEmbeddingsInterface(client=client,model_id="sentence-transformers/all-minilm-l6-v2",parameters=TextEmbeddingParameters(truncate_input_tokens=True))


# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
import os
from langchain_openai import ChatOpenAI
from langchain.embeddings.openai import OpenAIEmbeddings

from sax.core.synthesis.llms.base_llm import BaseLLM

class OpenAILLM(BaseLLM):
    """OpenAI LLM wrapper, implementation of BaseLLM
    """
    def __init__(self, model, temperature):
        self.api_key = os.getenv("OPENAI_KEY")
        if not self.api_key:
            raise ValueError("The API key for OpenAI is not set. Please set the OPENAI_KEY environment variable.")  
        super().__init__(model,temperature)         
        self._model = ChatOpenAI(             
            model_name=model,
            openai_api_key=self.api_key,
            temperature=temperature
        ) 
        self._embeddingModel = OpenAIEmbeddings(openai_api_key=self.api_key) 

    



# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
import os
import tempfile
from pathlib import Path

from flask import Blueprint, current_app, jsonify, make_response, request

from sax.api.exceptions.validation import ValidationException
from sax.core.causal_process_discovery import causal_discovery as cd
from sax.core.causal_process_discovery.causal_constants import Modality
from sax.core.process_mining import process_mining as pm
from sax.core.synthesis import sax_explainability as sx

sax_routes = Blueprint("sax", __name__, url_prefix="/sax")

def import_event_log_file(file,case_id,activity_key,timestamp_key,timestamp_format,lifecycle_type=None, start_timestamp_key=None,):          
    temp_dir = tempfile.gettempdir()
    temp_file_path = os.path.join(temp_dir, file.filename)
    file.save(temp_file_path)

    # Call the third-party method with the file path
    file_extension = Path(file.filename).suffix.lower()
    params = {"eventlog":temp_file_path,"case_id": case_id, "activity_key": activity_key, "timestamp_key": timestamp_key,"timestamp_format":timestamp_format}    
    if lifecycle_type is not None:
        params["lifecycle_type"]:lifecycle_type   
    if file_extension == '.csv':
        if start_timestamp_key is not None:
            params["starttime_column"]=start_timestamp_key
        dataframe = pm.import_csv(**params)
    elif file_extension == '.mxml':
        dataframe = pm.import_mxml(**params)
    elif file_extension == '.xes':
        dataframe = pm.import_xes(**params)   
    else:
        raise ValidationException("ValidationError: Incorrect file type, sax4bpm supports .xes, .mxml, .csv file types","Incorrect file type, sax4bpm supports .xes, .mxml, .csv file types")

    # Clean up the temporary file
    os.remove(temp_file_path)
    current_app.state.update_state(dataframe)

    return f"Finished importing event log file"

def get_variants():        
    imported_data = current_app.state.get_state_copy()   
    if not imported_data:
         raise ValidationException("ValidationError: No data imported yet. Call importEventLogFile first","No data imported yet. Call importEventLogFile first")        
    return imported_data.getVariants()

def get_process_model(variant_name, model_type):
    imported_data = current_app.state.get_state_copy()   
    if not imported_data:
         raise ValidationException("ValidationError: No data imported yet. Call importEventLogFile first","No data imported yet. Call importEventLogFile first")
    
    if variant_name != 'ALL':
        variant = imported_data.getVariant(variant_name)        
    else:
        variant = imported_data
    
    dfg = None
    if model_type == 'DFG':
        dfg,event_log = pm.discover_dfg(variant)        
    elif model_type == 'HEURISTIC':
        heuristic_net= pm.discover_heuristics_net(variant)       
        dfg = heuristic_net.dfg               
    else:
        raise ValidationException("ValidationError: Unsupported process model for discovery, use DFG or HEURISTIC"," Unsupported process model for discovery, use DFG or HEURISTIC")
    result_list =  [{"from": pair[0], "to": pair[1], "items": count} for pair, count in dfg.items()]
    return result_list

def get_causal_graph(variant_name, modality, prior_knowledge = False, p_value_threshold = None):
    imported_data = current_app.state.get_state_copy()   
    if not imported_data:
         raise ValidationException("ValidationError: No data imported yet. Call importEventLogFile first","No data imported yet. Call importEventLogFile first")
    
    if variant_name != 'ALL':
        variant = imported_data.getVariant(variant_name)        
    else:
        variant = imported_data   
    modality = Modality.from_string(modality)     

    causal_graph = cd.getDataCausalRepresentation(variant,modality,prior_knowledge,p_value_threshold)
    result_list =  [{"from": pair[0], "to": pair[1], "items": count} for pair, count in causal_graph.items()]
    return result_list

def get_explanations(variant_name,modality,prior_knowledge,p_threshold=None):   
    imported_data = current_app.state.get_state_copy()   
    if not imported_data:
         raise ValidationException("ValidationError: No data imported yet. Call importEventLogFile first","No data imported yet. Call importEventLogFile first")
    if variant_name != 'ALL':
        variant = imported_data.getVariant(variant_name)        
    else:
        variant = imported_data   
    modality = Modality.from_string(modality)     
    return sx.getExplanations(variant,modality=modality,prior_knowledge=prior_knowledge,p_value_threshold=p_threshold)

@sax_routes.route("/test_response")
def test_repsonse():
    return make_response(
        'Test worked!',
        200
    )


@sax_routes.route('/importEventLogFile', methods=['POST'])
def import_event_log_file_route():
    """
    Endpoint to import event log file.

    ---
    parameters:
    - name: file
      in: formData
      type: file
      required: true
      description: The event log file to import
    - name: case_id
      in: query
      type: string
      required: true
      description: Case ID column or attribute name
    - name: activity_key
      in: query
      type: string
      required: true
      description: Activity column or attribute name
    - name: timestamp_key
      in: query
      type: string
      required: true
      description: Timestamp column or attribute name
    - name: timestamp_format
      in: query
      type: string
      required: true
      description: Timestamp column or attribute name
    - name: lifecycle_type
      in: query
      type: string
      required: false
      description: Lifecycle column or attribute name if present in the event log
    - name: start_timestamp_key
      in: query
      type: string
      required: false
      description: Start Timestamp column or attribute name if present in the event log

    responses:
      200:
        description: Successfully imported event log file
      400:
        description: No event log part
      500:
        description: Internal Server Error
    """    
    case_id = request.args.get('case_id')
    activity_key = request.args.get('activity_key')
    timestamp_key = request.args.get('timestamp_key')
    timestamp_format = request.args.get('timestamp_format')
    lifecycle_type = request.args.get('lifecycle_type', None)
    start_timestamp_key = request.args.get('start_timestamp_key', None)
    
    if 'file' not in request.files:
        raise ValidationException("ValidationError:No event log attached to the request","No event log attached to the request")        
    
    file = request.files['file']
    
    if file.filename == '':
        raise ValidationException("ValidationError:No event log attached to the request","No event log attached to the request") 
    
    try:
      result = import_event_log_file(file,case_id,activity_key,timestamp_key,timestamp_format,lifecycle_type, start_timestamp_key)    
    except Exception as e:
      error = str(e)
      raise ValidationException(error,"Application exception") from e
    return jsonify({"result": result})


@sax_routes.route('/variants', methods=['GET'])
def get_variants_route():
    """
    Endpoint to get variants and their instance counts.

    ---
    responses:
      200:
        description: A dictionary containing variant names and their instance counts.
        content:
          application/json:
            schema:
              type: object
              example:
                variant1: 100
                variant2: 50
                variant3: 75
      500:
        description: Internal Server Error
    """    
    try:
      variants = get_variants()    
    except Exception as e:
      error = str(e)
      raise ValidationException(error,"Application exception") from e
    return jsonify(variants)

@sax_routes.route('/processModel', methods=['GET'])
def get_process_model_route():
    """
    Endpoint to get the process model for a given variant.

    ---
    parameters:
      - name: variant_name
        in: query
        type: string
        required: true
        description: The name of the variant (or ALL for all the variants)
      - name: model_type
        in: query
        type: string
        required: true
        description: The type of process model (e.g., DFG or HEURISTIC)

    responses:
      200:
        description: A JSON array containing the process model.
        content:
          application/json:
            schema:
              type: array
              items:
                type: object
                properties:
                  from:
                    type: string
                    description: The starting node of the transition.
                  to:
                    type: string
                    description: The ending node of the transition.
                  items:
                    type: integer
                    description: The number of occurrences of the transition.
      500:
        description: Internal Server Error
    """    
    variant_name = request.args.get('variant_name')
    model_type = request.args.get('model_type')    
    try:
      process_model = get_process_model(variant_name, model_type)    
    except Exception as e:
      error = str(e)
      raise ValidationException(error,"Application exception") from e
    return jsonify(process_model)

@sax_routes.route('/causalGraph', methods=['GET'])
def get_causal_graph_route():
    """
    Endpoint to get the causal graph for a given variant and modality.

    ---
    parameters:
      - name: variant_name
        in: query
        type: string
        required: true
        description: The name of the variant (or ALL for all the variants)
      - name: modality
        in: query
        type: string
        required: true
        description: The modality of causal graph ('Chain' or 'Parent')
      - name: prior_knowledge
        in: query
        type: boolean
        required: false
        description: Flag indicating whether prior knowledge should be used in the discovery process
      - name: p_value_threshold
        in: query
        type: number
        required: false
        description: The threshold value for p-value

    responses:
      200:
        description: A JSON array containing the causal graph.
        content:
          application/json:
            schema:
              type: array
              items:
                type: object
                properties:
                  from:
                    type: string
                    description: The starting node of the causal relationship.
                  to:
                    type: string
                    description: The ending node of the causal relationship.
                  items:
                    type: number
                    description: The strength of the causal relationship.
      500:
        description: Internal Server Error
    """

    variant_name = request.args.get('variant_name')
    modality = request.args.get('modality')
    prior_knowledge = request.args.get('prior_knowledge', None)
    p_value_threshold = request.args.get('p_value_threshold', None)
    try:
      causal_graph = get_causal_graph(variant_name, modality,prior_knowledge,p_value_threshold)            
    except Exception as e:
      error = str(e)
      raise ValidationException(error,"Application exception") from e
    return jsonify(causal_graph)

@sax_routes.route('/explanations', methods=['GET'])
def get_explanations_route():
    """
    Endpoint to get explanations for discrepancies between process and causal views in a given variant and modality.

    ---
    parameters:
      - name: variant_name
        in: query
        type: string
        required: true
        description: The name of the variant or ALL for the whole event log
      - name: modality
        in: query
        type: string
        required: true
        description: The modality for causal discovery ('Chain' or 'Parent')
      - name: prior_knowledge
        in: query
        type: boolean
        required: false
        description: Flag indicating whether prior knowledge should be used in the explanation process
      - name: p_value_threshold
        in: query
        type: number
        required: false
        description: The threshold value for p-value

    responses:
      200:
        description: A JSON array containing explanations for discrepancies.
        content:
          application/json:
            schema:
              type: array
              items:
                type: string
                description: Explanation for a discrepancy
      500:
        description: Internal Server Error
    """    
    variant_name = request.args.get('variant_name')
    modality = request.args.get('modality')
    prior_knowledge = request.args.get('prior_knowledge', None)
    p_threshold = request.args.get('p_value_threshold', None)    
    try:
      discrepancies = get_explanations(variant_name,modality,prior_knowledge,p_threshold)    
    except Exception as e:
      error = str(e)
      raise ValidationException(error,"Application exception") from e
    return jsonify(discrepancies)



# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
from enum import Enum

   
class Algorithm(Enum):
    """
    Enum representing causal discovery algorithms
    
    - LINGAM: DirectLINGAM algorithm
    """ 
    LINGAM = "Lingam"


class Modality(Enum):
    """
    Enum representing causal discovery modalities

    - PARENT: Parent anchor modality, where the examined time relationship between activities A->B->C is B-A and C-A (A is the parent activity and we examine the duration of B after A, and the duration of C which is combined durations of B after A and C after B)
    - CHAIN: Chain anchor modality, where the examined time relationships are between time intervals of each activity timestamp to the initial first activity timestamp
    """          
    PARENT = "Parent"
    CHAIN = "Chain"

    @classmethod
    def from_string(cls, value):
        for modality in cls:
            if value.lower() == modality.value.lower():
                return modality
        raise ValueError(f"No such modality: {value}")

DEFAULT_VARIANT = Algorithm.LINGAM
DEFAULT_MODALITY = Modality.CHAIN



# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
import pandas as pd

from sax.core.process_data.raw_event_data import RawEventData
from sax.core.utils import helper_utils
from sax.core.utils.constants import Constants, ConstantsMeta
from .base_formatter import BaseFormatter


class CSVFormatter(BaseFormatter):
    '''
    Formatter class for CSV event log files transforming the file to standard event log dataframe representation
    '''      

    class Parameters(metaclass=ConstantsMeta):
        """
        Constants representing various default values for mandatory attributes for CSV-based event log
        """
        ACTIVITY = "Source"
        CASE_ID = "Id"
        TYPE = "Type"
        TIMESTAMP = "Timestamp"
        TIMESTAMP_FORMAT = "%Y-%m-%d %H:%M:%S%z"
        RESOURCE_ID = "Resource"
        CSV_SEPARATOR = ","
        STARTTIME_COLUMN = ""
        

    def __init__(self,  parameters=None):
        """
        Initialize a new instance of the CSVFormatter class.

        Parameters
        ----------
        parameters : dict, optional
            A dictionary of parameters to use when formatting the event log, by default None.

        Raises
        ------
        ValueError
            If the parameters argument is not a dictionary or None.
        """
        super().__init__(parameters=parameters)
        self.parameters[Constants.ACTIVITY_KEY]= helper_utils.get_param_value(Constants.ACTIVITY_KEY, parameters, CSVFormatter.Parameters.ACTIVITY)
        self.parameters[Constants.CASE_ID_KEY] = helper_utils.get_param_value(Constants.CASE_ID_KEY, parameters, CSVFormatter.Parameters.CASE_ID)
        self.parameters[Constants.TIMESTAMP_KEY] = helper_utils.get_param_value(Constants.TIMESTAMP_KEY, parameters, CSVFormatter.Parameters.TIMESTAMP)
        self.parameters[Constants.TYPE_KEY] = helper_utils.get_param_value(Constants.TYPE_KEY, parameters, CSVFormatter.Parameters.TYPE)
        self.parameters[Constants.TIMESTAMP_FORMAT_KEY] = helper_utils.get_param_value(Constants.TIMESTAMP_FORMAT_KEY, parameters, CSVFormatter.Parameters.TIMESTAMP_FORMAT)
        self.parameters[Constants.STARTTIME_COLUMN]= helper_utils.get_param_value(Constants.STARTTIME_COLUMN, parameters, CSVFormatter.Parameters.STARTTIME_COLUMN)
        
    
    
    def extract_data(self,event_log_data,separator=Constants.CSV_SEPARATOR) -> RawEventData:
        """
        Extract tabular data from the provided CSV event log file, instantiate raw data object containing the tabular data.

        Parameters
        ----------
        event_log_data : file
            File representing the event log data, in CSV format.
        separator : str, optional
            The separator used in the CSV file, by default Constants.CSV_SEPARATOR.

        Returns
        -------
        RawEventData
            A RawEventData representing the event log in a raw event format.

        Raises
        ------
        ValueError
            If the event_log_data argument is not a file.
        """           
        dataframe = pd.read_csv(event_log_data, sep=separator) #original dataframe
        return self._format_dataframe(dataframe)
    
    def _format_dataframe(self, dataframe : pd.DataFrame) -> RawEventData:
        """
        Format the provided dataframe to represent event log: format timestamps, calculate trace start timestamps

        Parameters
        ----------
        dataframe : pd.DataFrame
            The dataframe to be formatted.

        Returns
        -------
        RawEventData
            A RawEventData object containing the formatted data.

        Raises
        ------
        ValueError
            If the dataframe argument is not a pandas DataFrame.
        """
        original_data = pd.DataFrame(dataframe)        
        mandatory_properties,optional_properties = self._getProperties(original_data, self.parameters)
        if Constants.STARTTIME_COLUMN in mandatory_properties: 
            #calculate the time for each trace based on this column, add it as separate column
            event_log = helper_utils.convert_timestamp_columns_in_df(original_data, timest_format=self.parameters[Constants.TIMESTAMP_FORMAT_KEY], timest_columns=[self.parameters[Constants.TIMESTAMP_KEY], self.parameters[Constants.STARTTIME_COLUMN]])
            event_log = helper_utils.add_start_time(event_log, timestamp_column_name=mandatory_properties[Constants.STARTTIME_COLUMN], id_column_name=mandatory_properties[Constants.CASE_ID_KEY], start_column_name=Constants.STARTTIME_COLUMN)
            #override to this column to be the indicator for trace start time
            mandatory_properties[Constants.STARTTIME_COLUMN]=Constants.STARTTIME_COLUMN  
        else:
            #create such a column for each row and add it to mandatory properties
            event_log = helper_utils.convert_timestamp_columns_in_df(original_data, timest_format=self.parameters[Constants.TIMESTAMP_FORMAT_KEY], timest_columns=[self.parameters[Constants.TIMESTAMP_KEY]])
            event_log = helper_utils.add_start_time(event_log, timestamp_column_name=mandatory_properties[Constants.TIMESTAMP_KEY], id_column_name=mandatory_properties[Constants.CASE_ID_KEY], start_column_name=Constants.STARTTIME_COLUMN)
            mandatory_properties[Constants.STARTTIME_COLUMN]=Constants.STARTTIME_COLUMN                                                                        
        return RawEventData(event_log,mandatory_properties,optional_properties)
    
    def extract_from_dataframe(self,dataframe: pd.DataFrame) -> RawEventData:
        """
        Extract event log tabular data from the provided dataframe, instantiate process raw event data object.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Dataframe representing the event log data.

        Returns
        -------
        RawEventData
            A RawEventData representing the event log in a raw event format.

        Raises
        ------
        ValueError
            If the dataframe argument is not a pandas DataFrame.
        """                  
        
        return self._format_dataframe(dataframe)


# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
from typing import Optional

import numpy as np
import pandas as pd

from sax.core.utils.constants import Constants
from ...causal_process_discovery.algorithms.base_causal_alg import CausalResultInfo
from ...causal_process_discovery.algorithms.lingam import LingamImpl
from ...causal_process_discovery.causal_constants import DEFAULT_VARIANT, Algorithm
from ...causal_process_discovery.modalities.base_anchor import BaseAnchor
from ...causal_process_discovery.prior_knowledge import PriorKnowledge
from ...process_data.raw_event_data import RawEventData


class ParentAnchorTransformer(BaseAnchor):
    '''
    Parent Causal Modality - methods to apply causal discovery on event log with PARENT modality
    '''    

    
    def apply(self, dataObject: RawEventData,variant: Optional[Algorithm] = DEFAULT_VARIANT,prior_knowledge: Optional[bool]=False, depth: int=1) -> CausalResultInfo:
        # Assisted by WCA for GP
        # Latest GenAI contribution: granite-20B-code-instruct-v2 model
        """
        Transpose the original event log so that each column is activity name and the value is the timestamp of the activity. Calculate all possible pairs of activities
        within this process and all its variants which happen one after the other. For each part, across all variants holding this pair, compute the time diff with the 
        parent activity of the pair. Apply causal discovery algorithm of the subframe defined by the parent activity and the pair of activities and the  time diff between them. 
        This will be a single coefficient in the adjacenty matrix of the whole process. At the end create global adjacency matrix out of all coefficients.

        Parameters
        ----------
        dataObject : RawEventData
            event log data object
        variant : Optional[Algorithm], optional
            algorithm variant to apply, defaults to LINGAM
        prior_knowledge : Optional[bool], optional
            whether to apply prior knowledge, defaults to False
        depth : int, optional
            depth of the distance between activities to consider as pair, defaults to 1

        Raises
        ------
        TypeError
            The method can be applied only to an object of type RawEventData!

        Returns
        -------
        CausalResultInfo
            causal discovery result
        """

        if type(dataObject) not in [RawEventData]: raise TypeError("the method can be applied only to an object of type RawEventData!")
       
        transposed = dataObject.transposeToTabular()    
        #change the id field name
        transposed_df = transposed.getData()
        caseColumnName = transposed.getMandatoryProperties()[Constants.CASE_ID_KEY]
        #rename the current column name to constant
        transposed_df.rename(columns={caseColumnName: Constants.CASE_ID_KEY}, inplace=True)

        #variants = dataObject.getVariants()
        # apply the name of the variant to each caseid
        variants_keys = dataObject.getVariantsKeys()
        transposed_df = self._join_keys(transposed_df,variants_keys)
        # iterate over the variants. va
        # For the first variant create the first global adjacency matrix
        tuples_map={}
        start_time_column_name = transposed.getMandatoryProperties()[Constants.STARTTIME_COLUMN]
        for key, value in variants_keys.items():
            
            #translate the key to list of activities
            activity_list = key.rstrip(",").split(",")
            #TODO: extend the implementation of _compute_tuples so that includes all permutations of forward facing activities
            tuples = self._compute_tuples(start_time_column_name,activity_list,depth) #return an array of tuples where each tuple consists of pair of consecutive activities and the parent activity of this pair
            for parent_activity, child_tuple in tuples:                
                reversed_tuple = tuple(reversed(child_tuple))

                if (child_tuple in tuples_map) or (reversed_tuple in tuples_map): #TODO need to check either order                    
                    continue
    
                # Accessing individual elements within the inner tuple
                first_activity = child_tuple[0]
                second_activity = child_tuple[1]
                if first_activity == second_activity:
                    #handle rework
                    continue                
                coefficient = self._apply_lingam(transposed_df,start_time_column_name,variant,first_activity, second_activity,prior_knowledge)
                tuples_map[child_tuple] = coefficient
        global_adj_matrix,list_columns= self._build_global_adj_matrix(tuples_map)
        #return result object built from global matrix
        return CausalResultInfo(global_adj_matrix,list_columns)
    
    def _get_variant(self,row,variant_map):
        # Assisted by WCA for GP
        # Latest GenAI contribution: granite-20B-code-instruct-v2 model
        """
        Get the process variant the given trace belongs to

        Parameters
        ----------
        row : pd.Series
            Row of the transposed data frame
        variant_map : dict
            Dictionary containing the variants and the caseIDs of the traces belonging to this variant

        Returns
        -------
        str
            Variant of the case
        """

        for variant, case_ids in variant_map.items():
            if row[Constants.CASE_ID_KEY] in case_ids:
                return variant
        return None

    def _join_keys(self,transposed_df,variants_keys)-> pd.DataFrame:    
        # Assisted by WCA for GP
        # Latest GenAI contribution: granite-20B-code-instruct-v2 model
        """
        For each trace in the dataframe compute which variant it belongs to and assign the name of the variant of each trace in the 'variant' column

        Parameters
        ----------
        transposed_df : pd.DataFrame
            Transposed data frame
        variants_keys : dict
            Dictionary containing the variants and their respective keys

        Returns
        -------
        pd.DataFrame
            Data frame with variant names for each trace
        """
   
        # Apply the _get_variant method to each row in transposed_df        
        transposed_df['variant'] = transposed_df.apply(lambda row: self._get_variant(row, variants_keys), axis=1)      
        return transposed_df
    

    
    def _compute_tuples(self,start_time_column,activity_list, depth):    
        # Assisted by WCA for GP
        # Latest GenAI contribution: granite-20B-code-instruct-v2 model
        """
        Compute all possible tuples of activities within the variant specified by list of activities where the distance between the activities is less or equal to given depth

        Parameters
        ----------
        start_time_column : str
            Name of the start time column
        activity_list : list
            List of activities
        depth : int
            Depth of the distance between tuples

        Returns
        -------
        list
            List of tuples of activities
        """

        result = []
        #adding start time column name as the first activitiy in the list for the purpose of calculating the pairs
        for i in range(len(activity_list) - 1):
            first_activity = activity_list[i]
            if (i == 0): parent_activity = start_time_column
            else : parent_activity = activity_list[i-1]
            j=0
            while (j < depth and i+j+1 <len(activity_list)):            
                second_activity = activity_list[i+j+1]
                child_tuple = (first_activity, second_activity)
                result.append((parent_activity, child_tuple))
                j+=1
        return result
    
    def _calculate_df(self, transposed_df, first_activity, second_activity, parent_activity):     
        # Assisted by WCA for GP
        # Latest GenAI contribution: granite-20B-code-instruct-v2 model
        """
        Given the dataframe, build a sub-dataframe composed only of the given three columns, and then calculate the time difference between the parent activity and each of the two activities 
        for each trace.

        Parameters
        ----------
        transposed_df : pandas.DataFrame
            Transposed dataframe containing the events
        first_activity : str
            First activity in the tuple
        second_activity : str
            Second activity in the tuple
        parent_activity : str
            Parent activity in the tuple

        Returns
        -------
        pandas.DataFrame
            Dataframe containing the time differences
        """
   
        subset_df = transposed_df[[parent_activity, first_activity, second_activity]]
        subset_df_clean = subset_df.apply(pd.to_datetime, utc=True) 
        first_activity_time = subset_df_clean.min(axis=1)      
   
        # Create a new DataFrame for storing time differences
        time_difference_df = subset_df_clean.copy()
        for col in subset_df_clean.columns:
            time_difference_df[col] = (subset_df_clean[col] - first_activity_time).dt.total_seconds()*1000
        return time_difference_df

    def _get_parent_string(self, start_time_column, string_list, first_activity,second_activity):
        # Assisted by WCA for GP
        # Latest GenAI contribution: granite-20B-code-instruct-v2 model
        """
        Get the parent string for a given pair of activities - activity which happens just before the earlies activity of the two. If there is no such activity
        (the given activities are the first in the process) then the start time is the "parent" activity

        Parameters
        ----------
        start_time_column : str
            name of the start time column
        string_list : list
            list of strings representing activities
        first_activity : str
            first activity in the tuple
        second_activity : str
            second activity in the tuple

        Returns
        -------
        str
            parent string
        """

        #need to accept also the start time
        #need to examine if first activity is the first activity in the list, and if so then the parent activity shoud be start time
        #there is a chance that right now this function will not be called for first activity..because parent is missing for such
        for i in range(len(string_list) - 1):
            if string_list[i] == first_activity or string_list[i] == second_activity:
                if i > 0: return string_list[i-1]
                else: return start_time_column
        return None  # Return None if the tuple is not found

    def _apply_lingam(self,transposed_df,start_time_column_name, alg_variant, first_activity,second_activity,prior_knowledge):  
        # Assisted by WCA for GP
        # Latest GenAI contribution: granite-20B-code-instruct-v2 model
        """
        Apply Lingam algorithm on the given sub-data frame specified by the first activity and second activity column names. The time difference is calculated on 
        per variant basis (for all variants which have those activities) vrs parent activity in this variant. After time diff calculations, the chosen causal discovery algorithm
        is applied on this dataframe to produce adjacency matrix

        Parameters
        ----------
        transposed_df : pd.DataFrame
            data frame containing the event log data
        start_time_column_name : str
            name of the start time column - the column holding start time of each process case
        alg_variant : Algorithm
            algorithm variant to apply
        first_activity : str
            name of the first activity
        second_activity : str
            name of the second activity
        prior_knowledge : bool
            whether to apply prior knowledge

        Returns
        -------
        float
            coefficient of the causal relationship between the two activities
        """
        
        # retrieve a sub df with all variants where those activities are not NaN
        subset_df_clean = transposed_df.dropna(subset=[first_activity, second_activity])
        # retrive a list of variants for those cases
        result_list = subset_df_clean['variant'].unique()
        result_df = pd.DataFrame()
        # iterate over the variants
        for variant in result_list:
        # for each variants retrieve the name of parent activity (the activity preceeding the first activity or the second activity if the order is opposite) 
            activity_list = variant.rstrip(",").split(",")
            parent_activity = self._get_parent_string(start_time_column_name,activity_list,first_activity,second_activity)
            if parent_activity is None: continue         
            #take the variant sub df from the overall transposed df
            subset_variant_df = subset_df_clean[subset_df_clean['variant'] == variant]   
            # calculate sub df given the parent activity and the two activities columns                      
            new_df = self._calculate_df(subset_variant_df,first_activity,second_activity,parent_activity)
            new_column_names = ['Parent', first_activity, second_activity]
            new_df.columns = new_column_names
            # combine all dfs and apply lingam
            result_df = pd.concat([result_df, new_df], ignore_index=True)        
        # retrieve coefficient for our main columns
        num_rows = result_df.shape[0]
        #Given the two columns of time difference, apply Lingam    
        args = {"data": result_df}
        if prior_knowledge:
            args["prior_knowledge"] = PriorKnowledge(result_df)
        if num_rows > 0: 
            if alg_variant == Algorithm.LINGAM:
                algorithm = LingamImpl(**args)
                try:
                    result  = algorithm.run()     
                    adjacency_matrix = result.getAdjacencyMatrix()
                    return adjacency_matrix[2,1]             
                except Exception as e:
                    print(e)                                
        return 0
         

 
    def _build_global_adj_matrix(self, coefficient_dict):
        # Assisted by WCA for GP
        # Latest GenAI contribution: granite-20B-code-instruct-v2 model
        """
        Build a global adjacency matrix from a dictionary of coefficients.

        Parameters
        ----------
        coefficient_dict : dict
            dictionary of coefficients

        Returns
        -------
        np.ndarray
            global adjacency matrix
        list
            list of unique elements
        """

        # Step 1: Identify unique elements
        unique_elements = set()
        for key in coefficient_dict.keys():
            unique_elements.update(key)

        # Step 2: Create an empty matrix with dimensions NxN
        num_nodes = len(unique_elements)
        adjacency_matrix = np.zeros((num_nodes, num_nodes))

        # Step 3: Populate the matrix with coefficients
        element_to_index = {element: index for index, element in enumerate(unique_elements)}
  
        for key, value in coefficient_dict.items():
            i, j = element_to_index[key[0]], element_to_index[key[1]]
            adjacency_matrix[j, i] = value  # Swap i and j

        return adjacency_matrix, list(unique_elements)

    

    




# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
from typing import Optional

import graphviz
import numpy as np
from lingam.utils import make_dot

from sax.core.causal_process_discovery.algorithms.base_causal_alg import CausalResultInfo
from sax.core.causal_process_discovery.causal_constants import DEFAULT_MODALITY, DEFAULT_VARIANT, Algorithm, Modality
from sax.core.causal_process_discovery.modalities.chain_anchor import ChainAnchorTransformer
from sax.core.causal_process_discovery.modalities.parent_anchor import ParentAnchorTransformer
from sax.core.process_data.raw_event_data import RawEventData


def discover_causal_dependencies(dataObject:RawEventData,variant: Optional[Algorithm] = DEFAULT_VARIANT, modality: Optional[Modality] = DEFAULT_MODALITY,prior_knowledge: Optional[bool]=False, depth: int =1) -> CausalResultInfo:
    """
    Create causal execution dependency model for the given event log

    :param dataObject: event log
    :type dataObject: RawEventData
    :param variant: Algorithm to use for causal discovery, defaults to Lingam
    :type variant: Optional[Algorithm], optional
    :param modality: Anchor modality to use for discovery, defaults to CHAIN modality
    :type modality: Optional[Modality], optional
    :param prior_knowledge: whether to use prior knowledge, defaults to False
    :type prior_knowledge: Optional[bool], optional
    :raises TypeError: in case the event log is not of appropriate format
    :return: causal dependency model representation
    :rtype: CausalResultInfo
    """       
    # Perform causal discovery algorithm here and generate the dependencies            
    if type(dataObject) not in [RawEventData]: raise TypeError("the method can be applied only to an object of type RawEventData!")
    
    if modality == Modality.CHAIN:
        
        #check the required modality and invoke the relevant transformer with provided algorithm       
        result = ChainAnchorTransformer().apply(dataObject,variant,prior_knowledge)
    else : #modality parent
        result  = ParentAnchorTransformer().apply(dataObject,variant,prior_knowledge,depth) 

    return result     



def getDataCausalRepresentation(dataframe: RawEventData,modality,prior_knowledge,p_value_threshold):
        """
        The purpose of this function is to take a raw event log as input and output a dictionary representation of the causal model discovered from this event log.
        :param dataframe: A pandas dataframe containing the raw event log data.
        :type dataframe: RawEventData
        :return: A dictionary representing the causal model, where each key is a tuple representing a transition between two activities, and the value is the strength of that transition as determined causal discovery.
        :rtype: dict
        """
        causalModel = discover_causal_dependencies(dataObject=dataframe,modality=modality,prior_knowledge=prior_knowledge)
        causalPairs = getModelCausalRepresentation(causalModel,p_value_threshold)
        return causalPairs

def getModelCausalRepresentation(model,p_value_threshold):
    def _get_causal_graph_representation(result: CausalResultInfo,p_value_threshold=None):
        result_dict = {}
        np_matrix = result.getAdjacencyMatrix()
        if p_value_threshold is not None:
            np_matrix = np.array(np_matrix)

            # Create a boolean mask where True indicates that the value is below the p-value threshold
            mask = np_matrix < p_value_threshold 
            
            # Replace values below the p-value threshold with zeros
            np_matrix[mask] = 0

        for i, row in enumerate(np_matrix):
            for j, coefficient in enumerate(row):
                if coefficient != 0.0:
                    key = (result.columns[j], result.columns[i])
                    result_dict[key] = coefficient

        return result_dict
    causalRepresentation = _get_causal_graph_representation(model,p_value_threshold)
    return causalRepresentation

def view_causal_dependencies(dependencies: CausalResultInfo, p_value_threshold: float=None) ->graphviz.Digraph:
    """
    View the causal dependency model graph

    :param dependencies: causal dependency model representation
    :type dependencies: CausalResultInfo
    :param p_value_threshold: threshold for displaying causal relationships, no edges with coefficients less than specified will be displayed, defaults to None
    :type p_value_threshold: float, optional
    :return: graph
    :rtype: graphviz.Digraph
    """        
    # Convert the input matrix to a NumPy array for easier manipulation
    np_matrix = dependencies.getAdjacencyMatrix()
    if p_value_threshold is not None:
        np_matrix = np.array(np_matrix)

        # Create a boolean mask where True indicates that the value is below the p-value threshold
        mask = np_matrix < p_value_threshold 
        
        # Replace values below the p-value threshold with zeros
        np_matrix[mask] = 0
        
    return make_dot(np_matrix, labels = dependencies.getColumns())    



# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.output_parsers import StrOutputParser
from langchain_core.runnables import RunnableLambda, RunnableParallel
from sax.core.causal_process_discovery import causal_discovery as cd
from sax.core.causal_process_discovery.causal_constants import Modality
from sax.core.process_data.raw_event_data import RawEventData
from sax.core.process_mining import process_mining as pm
from sax.core.synthesis.llms.base_llm import BaseLLM, ModelTypes
from sax.core.synthesis.rag.base_retrieval import BaseRetriever
from sax.core.synthesis.rag.documents_retrieval import DocumentRetrieverLLM
from operator import itemgetter

def createDocumentContextRetriever(modelType:ModelTypes, modelName: str, temperature: int, documentsPath:str,filter=None,chunk_size=None,chunk_overlap=None,retrieved_results=None,dbPath=None) -> BaseRetriever:
    """Create a document context retriever to use later for retreiving query-appropriate context.

    :param modelType: the model provider type to use
    :type modelType: ModelTypes
    :param modelName: the name of the model by the model provider to use as generative model
    :type modelName: str
    :param temperature: the temperature to use with the generative model
    :type temperature: int
    :param documentsPath: the path to the document/documents' folder
    :type documentsPath: str
    :param filter: file extension of the files to be considered, defaults to None
    :type filter: str, optional
    :param chunk_size: chunk_size for document splitting, defaults to None
    :type chunk_size: int, optional
    :param chunk_overlap: chunk overlap for document splitting, defaults to None
    :type chunk_overlap: int, optional
    :param retrieved_results: number of results to retrieve, defaults to None
    :type retrieved_results: int, optional
    :param dbPath: path to the vectore store db, defaults to None
    :type dbPath: str, optional
    :return: document retriever to be used later in rag-enhanced syntethis
    :rtype: BaseRetriever
    """
    model = BaseLLM.getModelLLM(modelType, modelName, temperature)
    retriever = DocumentRetrieverLLM(model,documentsPath,filter,chunk_size=chunk_size,chunk_overlap=chunk_overlap,retrieved_results=retrieved_results,dbPath=dbPath)
    return retriever

def _getChain(model: BaseLLM,causal: bool, process:bool, xai:bool, rag:bool, retriever: DocumentRetrieverLLM =None):   
    """
    Based on the chosen perspectives, builds and returns Langchain chain to invoke later with the user-supplied query for explanation based on thoe chosen perspectives
    :param model: model wrapper
    :type model: BaseLLM
    :param causal: indication whether to use causal perspective
    :type causal: bool
    :param process: indication whether to use process perspective
    :type process: bool
    :param xai: indication whether to use xai perspective
    :type xai: bool
    :param rag: indicatin whether to use document context
    :type rag: bool
    :param retriever: in case rag is chosen, need to provide previously created retreiver, defaults to None
    :type retriever: DocumentRetrieverLLM, optional
    :return: Langchain chain
    :rtype: chain
    """
    print(f"_getChain: process: {process}, causal: {causal}, xai: {xai}, rag: {rag}, retriever: {retriever}")
    def getCausalPerspective(data, modality, prior_knowledge,p_value_threshold):
        result = cd.getDataCausalRepresentation(data,modality=modality,prior_knowledge=prior_knowledge,p_value_threshold=p_value_threshold) 
        print("Causal: ",result)
        return result
    
    def getProcessPerspective(data):
        result =  pm.getDataProcessRepresentation(data)
        print("Process:" , result)
        return result 
    
    def getXAIPerspective(data):
        #TODO
        result = ""
        print("XAI", result)
        return result
    
    def getDocuments(retriever):
        def format_docs(docs):
            return "\n\n ------------".join(doc.page_content for doc in docs)
        retriver = retriever.get_retriever()
        return retriver | format_docs
    
    causal_runnable = RunnableLambda(lambda x: {"causal": getCausalPerspective(x["data"],x["modality"],x["prior_knowledge"],x["p_value_threshold"])})
    process_runnable = RunnableLambda(lambda x: {"process": getProcessPerspective(x["data"])})
    xai_runnable= RunnableLambda(lambda x: {"xai": getXAIPerspective(x["data"])})
    if rag:
        rag_retriever = getDocuments(retriever)
    else:
        rag_retriever=None

    if (not process and not causal and not xai and rag ):
        print("Method for combination (False, False, False, True)")
        prompt = ChatPromptTemplate.from_template(
        """We have some relevant documentation for a particular business process. Given the query {query} regarding this business process please answer the query based on the documentation: {documentation}
        Please answer the query in light of this information about the business process. Provide a detailed answer.                        
        """        
        )
        chain3 = prompt | model.getModel() | StrOutputParser()

        combined_chain = (
            RunnableParallel(query=itemgetter("query"), documentation=RunnableLambda(lambda inputs: inputs["query"])|rag_retriever) | chain3
        )
        
        return combined_chain

    if (not process and not causal and xai and not rag ):
        print("Method for combination (False, False, True, False)")        
        prompt = ChatPromptTemplate.from_template(
        """We have aמ XAI explainability view for a particular business process. Given the query {query} regarding this business process please answer the query based on the following information:             
            XAI explainability view: {xai} It is a dictionary, the key in which is a string, representing an activity, and the value is another dictionary, where the key is a string, representing an attribute of the activity, and the value is a float, representing the importance of that attribute in explaining the query.            
            Please answer the query in light of this information about the business process. Provide a detailed answer.                        
        """)
        chain3 = prompt | model.getModel() | StrOutputParser()

        combined_chain = (
            RunnableParallel(query=itemgetter("query"), xai=xai_runnable) | chain3
        )
        
        return combined_chain        

    if (not process and not causal and  xai and rag ):
        print("Method for combination (False, False, True, True)")
        combine_answers_prompt = ChatPromptTemplate.from_template(
        """We have an XAI explainability view for a particular business process. We also have some relevant documentation for this business process. Given the query {query} regarding this business process please answer the query based on the following information:             
            XAI explainability view: {xai} It is a dictionary, the key in which is a string, representing an activity, and the value is another dictionary, where the key is a string, representing an attribute of the activity, and the value is a float, representing the importance of that attribute in explaining the query.
            Relevant documentation about the business process: {documentation}

            Please answer the query in light of this information about the business process. Provide a detailed answer.                        
        """        
        )
        chain3 = combine_answers_prompt | model.getModel() | StrOutputParser()

        combined_chain = (
            RunnableParallel(query=itemgetter("query"), documentation=RunnableLambda(lambda inputs: inputs["query"])|rag_retriever,xai=xai_runnable) | chain3    
        )
        return combined_chain

    if (not process and causal and not xai and not rag ):
        print("Method for combination (False, True, False, False)")
        combine_answers_prompt = ChatPromptTemplate.from_template(
        """We have a causal-dependency model for a particular business process. Given the query {query} regarding this business process please answer the query based on the following information:             
            Causal-dependency model  for this business process: {causal}. It is a dictionary, the key in which is a tuple with two strings, representing two activities A and B , so that A causes B, and the value is the connection strength between them.            
            Please answer the query in light of this information about the business process. Provide a detailed answer.                        
        """        
        )
        chain3 = combine_answers_prompt | model.getModel() | StrOutputParser()

        combined_chain = (
            RunnableParallel(query=itemgetter("query"), causal =causal_runnable) | chain3
        )
        return combined_chain

    if (not process and causal and not xai and rag ):
        print("Method for combination (False, True, False, True)")
        combine_answers_prompt = ChatPromptTemplate.from_template(
        """We have a causal-dependency model. We also have some relevant documentation for this business process. Given the query {query} regarding this business process please answer the query based on the following information:            
            Causal-dependency model  for this business process: {causal}. It is a dictionary, the key in which is a tuple with two strings, representing two activities A and B , so that A causes B, and the value is the connection strength between them.            
            Relevant documentation about the business process: {documentation}

            Please answer the query in light of this information about the business process. Provide a detailed answer.                        
        """        
        )
        chain3 = combine_answers_prompt | model.getModel() | StrOutputParser()

        combined_chain = (
            RunnableParallel(query=itemgetter("query"), documentation=RunnableLambda(lambda inputs: inputs["query"])|rag_retriever, causal =causal_runnable) | chain3
        )
        return combined_chain

    if (not process and causal and  xai and not rag ):
        print("Method for combination (False, True, True, False)")
        combine_answers_prompt = ChatPromptTemplate.from_template(
        """We have a causal-dependency model and XAI explainability view for a particular business process. Given the query {query} regarding this business process please answer the query based on the following information:             
            Causal-dependency model  for this business process: {causal}. It is a dictionary, the key in which is a tuple with two strings, representing two activities A and B , so that A causes B, and the value is the connection strength between them.
            XAI explainability view: {xai} It is a dictionary, the key in which is a string, representing an activity, and the value is another dictionary, where the key is a string, representing an attribute of the activity, and the value is a float, representing the importance of that attribute in explaining the query.

            Please answer the query in light of this information about the business process. Provide a detailed answer.                        
        """        
        )
        chain3 = combine_answers_prompt | model.getModel() | StrOutputParser()

        combined_chain = (
            RunnableParallel(query=itemgetter("query"), causal =causal_runnable, xai=xai_runnable) | chain3
        )
        return combined_chain
        

    if (not process and causal and  xai and rag ):
        print("Method for combination (False, True, True, True)")
        combine_answers_prompt = ChatPromptTemplate.from_template(
        """We have a causal-dependency model and XAI explainability view for a particular business process. We also have some relevant documentation for this business process. Given the query {query} regarding this business process please answer the query based on the following information:             
            Causal-dependency model  for this business process: {causal}. It is a dictionary, the key in which is a tuple with two strings, representing two activities A and B , so that A causes B, and the value is the connection strength between them.
            XAI explainability view: {xai} It is a dictionary, the key in which is a string, representing an activity, and the value is another dictionary, where the key is a string, representing an attribute of the activity, and the value is a float, representing the importance of that attribute in explaining the query.
            Relevant documentation about the business process: {documentation}

            Please answer the query in light of this information about the business process. Provide a detailed answer.                        
        """        
        )
        chain3 = combine_answers_prompt | model.getModel() | StrOutputParser()

        combined_chain = (
            RunnableParallel(query=itemgetter("query"), documentation=RunnableLambda(lambda inputs: inputs["query"])|rag_retriever, causal =causal_runnable, xai=xai_runnable) | chain3
        )
        return combined_chain

    if (process and not causal and not xai and not rag ):
        combine_answers_prompt = ChatPromptTemplate.from_template(
        """We have a process model for a particular business process. Given the query {query} regarding this business process please answer the query based on the following information: 
            Process Model for this business process: {process} . It is a dictionary, the key in which is a tuple with two strings, representing two activities A and B , so that A occurs before B, and the value is the connection strength (number of process traces where B follows A) between them.            

            Please answer the query in light of this information about the business process. Provide a detailed answer.                        
        """        
        )
        chain3 = combine_answers_prompt | model.getModel() | StrOutputParser()

        combined_chain = (
            RunnableParallel(query=itemgetter("query"), process=process_runnable) | chain3)

        return combined_chain

    if (process and not causal and not xai and rag ):
        print("Method for combination (True, False, False, True)")
        combine_answers_prompt = ChatPromptTemplate.from_template(
        """We have a process model for a particular business process. We also have some relevant documentation for this business process. Given the query {query} regarding this business process please answer the query based on the following information: 
            Process Model for this business process: {process} . It is a dictionary, the key in which is a tuple with two strings, representing two activities A and B , so that A occurs before B, and the value is the connection strength (number of process traces where B follows A) between them.            
            Relevant documentation about the business process: {documentation}

            Please answer the query in light of this information about the business process. Provide a detailed answer.                        
        """        
        )
        chain3 = combine_answers_prompt | model.getModel() | StrOutputParser()

        combined_chain = (
            RunnableParallel(query=itemgetter("query"), documentation=RunnableLambda(lambda inputs: inputs["query"])|rag_retriever,  process=process_runnable) | chain3)
        return combined_chain

    if (process and not causal and xai and not rag ):
        print("Method for combination (True, False, True, False)")
        combine_answers_prompt = ChatPromptTemplate.from_template(
        """We have a process model, and XAI explainability view for a particular business process. Given the query {query} regarding this business process please answer the query based on the following information: 
            Process Model for this business process: {process} . It is a dictionary, the key in which is a tuple with two strings, representing two activities A and B , so that A occurs before B, and the value is the connection strength (number of process traces where B follows A) between them.           
            XAI explainability view: {xai} It is a dictionary, the key in which is a string, representing an activity, and the value is another dictionary, where the key is a string, representing an attribute of the activity, and the value is a float, representing the importance of that attribute in explaining the query.

            Please answer the query in light of this information about the business process. Provide a detailed answer.                        
        """        
        )
        chain3 = combine_answers_prompt | model.getModel() | StrOutputParser()

        combined_chain = (
            RunnableParallel(query=itemgetter("query"), process=process_runnable,xai=xai_runnable) | chain3
        )
        
        return combined_chain

    if (process and not causal and  xai and rag ):
        print("Method for combination (True, False, True, True)")
        combine_answers_prompt = ChatPromptTemplate.from_template(
        """We have a process model and XAI explainability view for a particular business process. We also have some relevant documentation for this business process. Given the query {query} regarding this business process please answer the query based on the following information: 
            Process Model for this business process: {process} . It is a dictionary, the key in which is a tuple with two strings, representing two activities A and B , so that A occurs before B, and the value is the connection strength (number of process traces where B follows A) between them.            
            XAI explainability view: {xai} It is a dictionary, the key in which is a string, representing an activity, and the value is another dictionary, where the key is a string, representing an attribute of the activity, and the value is a float, representing the importance of that attribute in explaining the query.
            Relevant documentation about the business process: {documentation}

            Please answer the query in light of this information about the business process. Provide a detailed answer.                        
        """        
        )
        chain3 = combine_answers_prompt | model.getModel() | StrOutputParser()

        combined_chain = (
            RunnableParallel(query=itemgetter("query"), documentation=RunnableLambda(lambda inputs: inputs["query"])|rag_retriever, process=process_runnable,xai=xai_runnable) | chain3
        )
        return combined_chain

    if (process and causal and not xai and not rag ):
        print("Method for combination (True, True, False, False)")
        combine_answers_prompt = ChatPromptTemplate.from_template(
        """We have a process model and causal-dependency model for a particular business process.Given the query {query} regarding this business process please answer the query based on the following information: 
            Process Model for this business process: {process} . It is a dictionary, the key in which is a tuple with two strings, representing two activities A and B , so that A occurs before B, and the value is the connection strength (number of process traces where B follows A) between them.
            Causal-dependency model  for this business process: {causal}. It is a dictionary, the key in which is a tuple with two strings, representing two activities A and B , so that A causes B, and the value is the connection strength between them.                        

            Please answer the query in light of this information about the business process. Provide a detailed answer.                        
        """        
        )
        chain3 = combine_answers_prompt | model.getModel() | StrOutputParser()

        combined_chain = (
            RunnableParallel(query=itemgetter("query"), causal =causal_runnable, process=process_runnable) | chain3)

        return combined_chain

    if (process and causal and not xai and rag ):
        print("Method for combination (True, True, False, True)")
        combine_answers_prompt = ChatPromptTemplate.from_template(
        """We have a process model and causal-dependency model  for a particular business process. We also have some relevant documentation for this business process. Given the query {query} regarding this business process please answer the query based on the following information: 
            Process Model for this business process: {process} . It is a dictionary, the key in which is a tuple with two strings, representing two activities A and B , so that A occurs before B, and the value is the connection strength (number of process traces where B follows A) between them.
            Causal-dependency model  for this business process: {causal}. It is a dictionary, the key in which is a tuple with two strings, representing two activities A and B , so that A causes B, and the value is the connection strength between them.            
            Relevant documentation about the business process: {documentation}

            Please answer the query in light of this information about the business process. Provide a detailed answer.                        
        """        
        )
        chain3 = combine_answers_prompt | model.getModel() | StrOutputParser()

        combined_chain = (
            RunnableParallel(query=itemgetter("query"), documentation=RunnableLambda(lambda inputs: inputs["query"])|rag_retriever, causal =causal_runnable, process=process_runnable) | chain3)
        return combined_chain

    if (process and causal and  xai and not rag ):
        print("Method for combination (True, True, True, False)")
        combine_answers_prompt = ChatPromptTemplate.from_template(
        """We have a process model, causal-dependency model and XAI explainability view for a particular business process.  Given the query {query} regarding this business process please answer the query based on the following information: 
            Process Model for this business process: {process} . It is a dictionary, the key in which is a tuple with two strings, representing two activities A and B , so that A occurs before B, and the value is the connection strength (number of process traces where B follows A) between them.
            Causal-dependency model  for this business process: {causal}. It is a dictionary, the key in which is a tuple with two strings, representing two activities A and B , so that A causes B, and the value is the connection strength between them. Tuples for activities which appear in process model but do not appear in causal dependency model might mean that there is precedence relationship between the activities in the tuple, but no causal relationship.
            XAI explainability view: {xai} It is a dictionary, the key in which is a string, representing an activity, and the value is another dictionary, where the key is a string, representing an attribute of the activity, and the value is a float, representing the importance of that attribute in explaining the query.            

            Please answer the query in light of this information about the business process. Provide a detailed answer.                        
        """        
        )
        chain3 = combine_answers_prompt | model.getModel() | StrOutputParser()

        combined_chain = (
            RunnableParallel(query=itemgetter("query"), causal =causal_runnable, process=process_runnable,xai=xai_runnable) | chain3)
        return combined_chain

    if (process and causal and  xai and rag ):
        print("Method for combination (True, True, True, True)")
        combine_answers_prompt = ChatPromptTemplate.from_template(
        """We have a process model, causal-dependency model and XAI explainability view for a particular business process. We also have some relevant documentation for this business process. Given the query {query} regarding this business process please answer the query based on the following information: 
            Process Model for this business process: {process} . It is a dictionary, the key in which is a tuple with two strings, representing two activities A and B , so that A occurs before B, and the value is the connection strength (number of process traces where B follows A) between them.
            Causal-dependency model  for this business process: {causal}. It is a dictionary, the key in which is a tuple with two strings, representing two activities A and B , so that A causes B, and the value is the connection strength between them.
            XAI explainability view: {xai} It is a dictionary, the key in which is a string, representing an activity, and the value is another dictionary, where the key is a string, representing an attribute of the activity, and the value is a float, representing the importance of that attribute in explaining the query.
            Relevant documentation about the business process: {documentation}

            Please answer the query in light of this information about the business process. Provide a detailed answer.                        
        """        
        )
        chain3 = combine_answers_prompt | model.getModel() | StrOutputParser()

        combined_chain = (
            RunnableParallel(query=itemgetter("query"), documentation=RunnableLambda(lambda inputs: inputs["query"])|rag_retriever, causal =causal_runnable, process=process_runnable,xai=xai_runnable) | chain3)
        return combined_chain    

    

def getSyntethis(data:RawEventData, query:str, model: BaseLLM,causal: bool, process:bool, xai:bool, rag:bool, retriever: DocumentRetrieverLLM =None, modality =Modality.PARENT, prior_knowledge=True, p_value_threshold=None):
    """Get blended answer based on the user-chosen perspectives for the user provided query

    :param data: event log data
    :type data: RawEventData
    :param query: user query
    :type query: str
    :param model: LLM model wrapper
    :type model: BaseLLM
    :param causal: indication whether to use causal perspective
    :type causal: bool
    :param process: indication whether to use process perspective
    :type process: bool
    :param xai: indication whether to use xai perspective
    :type xai: bool
    :param rag: indicatin whether to use document context
    :type rag: bool
    :param retriever: in case rag is chosen, need to provide previously created retreiver, defaults to None
    :type retriever: DocumentRetrieverLLM, optional
    :param modality: in case causal perspective is chosen, the desired analysis modality, defaults to Modality.PARENT
    :type modality: Modality, optional
    :param prior_knowledge: in case causal perspective is chosen, whether to use prior knowledge or not, defaults to True
    :type prior_knowledge: bool, optional
    :param p_value_threshold: in case causal perspective is chosen, whether to filter causal connections below specified threshold, defaults to None
    :type p_value_threshold: float, optional
    :return: Blended anwer to the query based on all chosen perspectives
    :rtype: str
    """
    chain = _getChain(model,causal, process, xai, rag, retriever)
    result = chain.invoke({"data":data,"query":query,"modality":modality,"prior_knowledge":prior_knowledge,"p_value_threshold":p_value_threshold})
    return result

def getModel(modelType:ModelTypes, modelName: str, temperature: int):
    """Retrieve desired model wrapper based on the specified model provider, model name and temperature

    :param modelType: model provider type
    :type modelType: ModelTypes
    :param modelName: the name of the generative model provided by chosen provider
    :type modelName: str
    :param temperature: temperature to use for generation
    :type temperature: int
    :return: Chosen model wrapper
    :rtype: BaseLLM
    """

    return BaseLLM.getModelLLM(modelType, modelName, temperature)



def getExplanations(data:RawEventData,modality,prior_knowledge = None, p_value_threshold=None ):
    """Return an array of semantic explanations for all process-causal disrepancies 

    :param data: event log data
    :type data: RawEventData
    :param modality: chosen modality for causal discovery
    :type modality: Modality
    :param prior_knowledge: whether to use prior knowledge in causal discovery, defaults to None
    :type prior_knowledge: boolean, optional
    :param p_value_threshold: a filter for causal connections, connections with strength below the specified threshold will be disregarded, defaults to None
    :type p_value_threshold: int, optional
    :return: array of disrepancies
    :rtype: array[str]
    """
    dfg,event_log = pm.discover_dfg(data)  
    causalModel = cd.discover_causal_dependencies(dataObject=data,modality=modality,prior_knowledge=prior_knowledge)
    result = enumerateDisrepancies(dfg, causalModel,p_value_threshold=p_value_threshold)
    return result
      
def enumerateDisrepancies(processModel, causalModel,p_value_threshold=None):   
    """Enumerate the disrepancies between provided process and causal models

    :param processModel: process model as dfg dictionary
    :type processModel: dict
    :param causalModel: causal model as dictionary
    :type causalModel: dict
    :param p_value_threshold: filtering threshold for causal connections
    :type p_value_threshold: int, optional
    :return: array of disrepancies
    :rtype: array[str]
    """
    def _calculateDiff(processRepresentation,causalModel):   
        set1 = set(processRepresentation.keys())
        set2 = set(causalModel.keys())

        removed_edges = list(set1 - set2)
        addedEdges = list(set2 - set1)
        
        return addedEdges, removed_edges
    addedEdge = ' Altering the \'{first_activity}\' completion time is likely to affect the lead time of \'{second_activity}\''
    removedEdge = 'Altering the \'{first_activity}\' completion time is not likely to affect lead time of \'{second_activity}\' '    

    explanations = []
    processRepresentation = pm.getModelProcessRepresentation(processModel)
    causalModel = cd.getModelCausalRepresentation(causalModel,p_value_threshold=p_value_threshold)
    #calculateDiff should return tuples of activities in correct order between which edges were added or removed
    addedEdges,removedEdges = _calculateDiff(processRepresentation,causalModel)
    for i, pair in enumerate(addedEdges):
        first_activity = pair[0]
        second_activity = pair[1]
        explanation = addedEdge.format(first_activity=first_activity, second_activity=second_activity)
        explanations.append(explanation)

    for i, pair in enumerate(removedEdges):
        first_activity = pair[0]
        second_activity = pair[1]
        explanation = removedEdge.format(first_activity=first_activity, second_activity=second_activity)
        explanations.append(explanation)
    
    return explanations



# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
from enum import Enum

class ModelTypes(Enum):
    """Enum for supported model providers. Currently supported types are OPENAI and BAM (IBM WatsonX)
    """
    OPENAI = "openai"
    BAM = "bam"    

class BaseLLM:         
    """BaseLLM is a base class for all LLMs wrappers. It provides basic functionality such as setting the model name and temperature and returning the wrapper for the chosen model provider.
    After creating the wrapper can use it to get the pre-configured models for different tasks (generation/embedding)
  
    """
    def __init__(self,model_name, temperature):
       self._model_name = model_name
       self._temperature = temperature
       self._model= None
       self._embeddingModel = None

    
    def getModel(self):
        """Get the preconfigured generation model

        :return: generation model
        :rtype: Langchain generation model
        """
        return self._model
    
    def getEmbeddingModel(self):
        """Get the preconfigured embedding model

        :return: embdedding model
        :rtype: Langchain embedding model
        """
        return self._embeddingModel
    
    @staticmethod
    def getModelLLM(modelType: ModelTypes, modelName, temperature)-> 'BaseLLM':
        """
        Return the LLM wrapper of the chosen model provider type, with the chosen model name and temperature
        :param modelType: model provider type
        :type modelType: ModelTypes
        :param modelName: the fully qualified name of the generation model which will be supplied by this wrapper
        :type modelName: str
        :param temperature: the temperature for the generation model
        :type temperature: int
        :raises ValueError: incorrect model type specified
        :return: Exception
        :rtype: BaseLLM
        """
        if modelType == ModelTypes.OPENAI:
            from sax.core.synthesis.llms.openai import OpenAILLM
            return OpenAILLM(modelName,temperature)
        elif modelType == ModelTypes.BAM:
            from sax.core.synthesis.llms.watsonx import WatsonXLLM
            return WatsonXLLM(modelName,temperature)
        else:
            raise ValueError("Invalid type of model: ",modelType)
    


# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
from typing import Any, Dict

from pandas import DataFrame

from sax.core.utils.constants import Constants
from .data import BaseProcessDataObject


class TabularEventData(BaseProcessDataObject):
    """
    Tabular representation of event log, where each row represents a full trace,each column representing a particular activity within the trace. The column names
    are the names of the activities, the column values are the timestamps of activity completion times. Additional columns represent activity attribute, each attribute column name composed of
    ``<activityName>__<attributeName>``

    """  
  
    def __init__(self,data:DataFrame, mandatory_properties:dict,optional_properties:dict):       
        """
        Initialize a new instance of TabularEventData class.

        :param data: Dataframe containing the event log data
        :type data: DataFrame
        :param mandatory_properties: Dictionary containing the mandatory properties of the event log
        :type mandatory_properties: dict
        :param optional_properties: Dictionary containing the optional properties of the event log
        :type optional_properties: dict     
        """  
        super().__init__(data=data,mandatory_properties=mandatory_properties,optional_properties= optional_properties)

    def getCaseAndActivitiesData(self)->DataFrame:
        """
        Return the mandatory properties columns content
        Parameters
        ----------
        None

        Returns
        -------        
        :return: dataframe comprised only of the content of mandatory columns (caseID and activities columns)
        :rtype: DataFrame
        """        
        return self.getData()[list(self.getMandatoryProperties().values())]  
    
    def getCaseIDData(self)->DataFrame:
        """
        Return the values of case Id colum

        Parameters
        ----------
        None
                
        Returns
        ------- 
        :return: caseId column content
        :rtype: DataFrame
        """        
        return self.getData()[self.mandatory_properties[Constants.CASE_ID_KEY]]
    
    def getActivitiesData(self) ->DataFrame:
        """
        Return the content of activities columns 

        Parameters
        ----------
        None
                
        Returns
        ------- 

        :return: activities columns content
        :rtype: DataFrame
        """        
        return self.getCaseAndActivitiesData().drop(self.mandatory_properties[Constants.CASE_ID_KEY],axis=1) 
    
    def getActivitiesAttributesData(self)->DataFrame:
        """
        Return all optional attributes columns content
        
        Parameters
        ----------
        None
                
        Returns
        ------- 

        :return: optional attribute columns content
        :rtype: DataFrame
        """        
        return self.getData()[list(self.getOptionalProperties().values())]

    def getActivityAttributesData(self,activityName) -> DataFrame:
        """
        Return attribute columns for the specified activity

        Parameters
        ----------
        :param activityName: activityName
        :type activityName: str
        
        Returns
        ------- 
        
        :return: content of attribute columns for the specified activity
        :rtype: DataFrame
        """        
        activitiesData = self.getActivitiesAttributesData()
        matching_columns = [col for col in activitiesData.columns if col.startswith(activityName)]
        return self.getData()[list(matching_columns)] 
    
    
    def getTrace(self,pid:str) -> Dict[str, Any]:
        """
        Get a trace by its process id (case id).

        Parameters
        ----------
        pid : str
            Case id of the trace to retrieve.
            
        Returns
        ------- 
        Dict[str, Any]
            The trace data as a dictionary (activity names and timestamps)
        
        Raises
        ------
        KeyError
            If a trace with such case id does not exist in the log.
        """    
        return self._get_row_by_id(pid)
   
    
    def _get_row_by_id(self, unique_id):
        """
        Get a trace from the event log by its case id.

        Parameters
        ----------
        unique_id : str
            Case id of the row to retrieve.
            
        Returns
        ------- 
        Dict[str, Any]
            The row data as a dictionary.
        
        Raises
        ------
        KeyError
            If the case id does not exist in the log.
        """    
        # Filter the dataframe based on the given Id
        row = self.data[self.data[self.mandatory_properties[Constants.CASE_ID_KEY] ] == unique_id]        
        # Convert the row to a dictionary and return
        return row.to_dict(orient='records')[0] if not row.empty else None

        


# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------

class LLMApiErrorException(Exception):
   """Exception raised when an error occurs while calling an external LLM API"""
   def __init__(self, message, details):
        self.message = message
        self.details = details
        super().__init__(self.message)


# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
from enum import Enum


class ConstantsMeta(type):
    def __call__(cls, *args, **kwargs):
        raise TypeError("Cannot instantiate Constants class")
    def __setattr__(cls, name, value):
        raise AttributeError("Cannot modify constant values")

class LifecycleTypes(Enum):
    START = "start"
    ASSIGN = "assign"
    COMPLETE = "complete"  
    
class Constants(metaclass=ConstantsMeta):
    ACTIVITY_KEY = "concept:name"
    CASE_ID_KEY = "case:concept:name"
    TYPE_KEY = "lifecycle:transition"
    GROUP_KEY="Group"
    TIMESTAMP_KEY = "time:timestamp"
    TIMESTAMP_FORMAT_KEY = "TimestampFormat"
    RESOURCE_KEY = "org:resource"
    STARTTIME_COLUMN="start:timestamp"

    CSV_SEPARATOR=","    



# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
import graphviz
import numpy as np

from pandas import DataFrame


class PriorKnowledge:
    """
    Prior knowledge representation for a particular event log. This object should be created and passed to the causal dependency discovery algorithm in case the user intends
    to run the algorithm with prior knowledge. The rules for creation of prior knowledge matrix are described here https://lingam.readthedocs.io/en/stable/tutorial/pk_direct.html
    """    
    def __init__(self,data: DataFrame):
        '''
        This class allows adding any prior knoweledge to the causal discovery process 
        '''
        self.prior_knowledge = self._assesPriorKnowledge(data)
        self.data = data

    def _assesPriorKnowledge(self,data: DataFrame):      
        # Assisted by WCA for GP
        # Latest GenAI contribution: granite-20B-code-instruct-v2 model
        '''
        Asses prior knowledge of the provided event log
        
        Parameters
        ----------
        data : pandas.DataFrame
            Event log data

        Returns
        -------
        np.ndarray
            Prior knowledge matrix
        '''
        prior_knowledge = np.eye(len(data.columns)) -1
        for ind1, column1 in enumerate(data.columns):
            for ind2, column2 in enumerate(data.columns):
                if column1!=column2 :
                    residuals = data[column1] - data[column2]
                    if (residuals >= 0).all():                        
                        prior_knowledge[ind2][ind1] = 0
        return prior_knowledge
    
    def getPriorKnowledge(self):
        """
        Get the prior knowledge matrix embedded in this object

        Returns
        -------
        np.ndarray
            Prior knowledge matrix
        """        
        return self.prior_knowledge
    
    def make_prior_knowledge_graph(self):
        """
        # Assisted by WCA for GP
        # Latest GenAI contribution: granite-20B-code-instruct-v2 model
        Creates a graphviz.Digraph object representing the prior knowledge embedded in this object.

        Parameters
        ----------
        None

        Returns
        -------
        graphviz.Digraph
            Graphviz Digraph object representing the prior knowledge

        """        
        prior_knowledge_matrix = self.prior_knowledge
        d = graphviz.Digraph(engine='dot')

        labels = self.data.columns
        for label in labels:
            d.node(label, label)

        dirs = np.where(prior_knowledge_matrix > 0)
        for to, from_ in zip(dirs[0], dirs[1]):
            d.edge(labels[from_], labels[to])

        dirs = np.where(prior_knowledge_matrix < 0)
        for to, from_ in zip(dirs[0], dirs[1]):
            if to != from_:
                d.edge(labels[from_], labels[to], style='dashed')
        return d
       
    


# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
class LLMApiConnectionException(Exception):
   """Exception risen when there is connection problem to external LLM
   """
   def __init__(self, message, details):
        self.message = message
        self.details = details
        super().__init__(self.message)


# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
from typing import Tuple

import pm4py
from pm4py.algo.discovery.dfg import algorithm as dfg_discovery
from pm4py.objects.bpmn.obj import BPMN
from pm4py.objects.heuristics_net.obj import HeuristicsNet
from pm4py.objects.process_tree.obj import ProcessTree
from pm4py.visualization.dfg import visualizer as dfg_visualization

from sax.core.process_data.formatters.csv_formatter import CSVFormatter
from sax.core.process_data.formatters.mxml_formatter import MXMLFormatter
from sax.core.process_data.formatters.xes_formatter import XESFormatter
from sax.core.process_data.raw_event_data import RawEventData
from sax.core.utils.constants import Constants


def import_xes(eventlog,case_id: str=XESFormatter.Parameters.CASE_ID, activity_key: str=XESFormatter.Parameters.ACTIVITY, timestamp_key: str=XESFormatter.Parameters.TIMESTAMP, lifecycle_type: str= XESFormatter.Parameters.TYPE,timestamp_format: str=XESFormatter.Parameters.TIMESTAMP_FORMAT) ->RawEventData:
        """
        Parse XES file into event log

        Parameters
        ----------
        :param eventlog: XES event log file
        :type eventlog: Path to the file
        :param case_id: name of the case id column, defaults to XESFormatter.Parameters.CASE_ID
        :type case_id: str, optional
        :param activity_key: name of the activity column, defaults to XESFormatter.Parameters.ACTIVITY
        :type activity_key: str, optional
        :param timestamp_key: name of the timestamp column, defaults to XESFormatter.Parameters.TIMESTAMP
        :type timestamp_key: str, optional
        :param lifecycle_type: name of the event lifecycle column, defaults to XESFormatter.Parameters.TYPE
        :type lifecycle_type: str, optional
        :param timestamp_format: timestamp format, defaults to XESFormatter.Parameters.TIMESTAMP_FORMAT
        :type timestamp_format: str, optional
        :return: Raw event data object
        :rtype: RawEventData
        
        Raises:
        FileNotFoundError: If the specified event log file does not exist, this exception will be raised.
        """
        
        parameters = {}
        parameters[Constants.CASE_ID_KEY]=case_id
        parameters[Constants.ACTIVITY_KEY]=activity_key
        parameters[Constants.TIMESTAMP_KEY]=timestamp_key
        parameters[Constants.TIMESTAMP_FORMAT_KEY] = timestamp_format  
        parameters[Constants.TYPE_KEY]=lifecycle_type        
        formatter = XESFormatter(parameters)
        dataframe = formatter.extract_data(eventlog)
        return dataframe
    
   
def import_csv(eventlog,case_id: str=CSVFormatter.Parameters.CASE_ID, activity_key: str=CSVFormatter.Parameters.ACTIVITY, timestamp_key: str=CSVFormatter.Parameters.TIMESTAMP,lifecycle_type: str= CSVFormatter.Parameters.TYPE, timestamp_format: str=CSVFormatter.Parameters.TIMESTAMP_FORMAT, starttime_column: str=CSVFormatter.Parameters.STARTTIME_COLUMN) ->RawEventData:
        """
        Parse CSV file into event log

        Parameters
        -----------
        :param eventlog: CSV event log file
        :type eventlog: Path to the file
        :param case_id: name of the case id column, defaults to XESFormatter.Parameters.CASE_ID
        :type case_id: str, optional
        :param activity_key: name of the activity column, defaults to XESFormatter.Parameters.ACTIVITY
        :type activity_key: str, optional
        :param timestamp_key: name of the timestamp column, defaults to XESFormatter.Parameters.TIMESTAMP
        :type timestamp_key: str, optional
        :param lifecycle_type: name of the event lifecycle column, defaults to XESFormatter.Parameters.TYPE
        :type lifecycle_type: str, optional
        :param timestamp_format: timestamp format, defaults to XESFormatter.Parameters.TIMESTAMP_FORMAT
        :type timestamp_format: str, optional

        Returns
        -------
        :return: Raw event data object
        :rtype: RawEventData

        Raises:
        FileNotFoundError: If the specified event log file does not exist, this exception will be raised.
        """        
        parameters = {}
        parameters[Constants.CASE_ID_KEY]=case_id
        parameters[Constants.ACTIVITY_KEY]=activity_key
        parameters[Constants.TIMESTAMP_KEY]=timestamp_key
        parameters[Constants.TIMESTAMP_FORMAT_KEY] = timestamp_format    
        parameters[Constants.TYPE_KEY]=lifecycle_type      
        parameters[Constants.STARTTIME_COLUMN]=starttime_column
        formatter = CSVFormatter(parameters)
        dataframe = formatter.extract_data(eventlog)
        return dataframe
    
   
def create_from_dataframe(dataframe,case_id: str=CSVFormatter.Parameters.CASE_ID, activity_key: str=CSVFormatter.Parameters.ACTIVITY, timestamp_key: str=CSVFormatter.Parameters.TIMESTAMP, lifecycle_type: str= CSVFormatter.Parameters.TYPE,timestamp_format: str=CSVFormatter.Parameters.TIMESTAMP_FORMAT,starttime_column: str=CSVFormatter.Parameters.STARTTIME_COLUMN)->RawEventData:
        """
        Creates event log from dataframe
        
        Parameters
        -----------
        :param eventlog: XES event log file
        :type eventlog: Path to the file
        :param case_id: name of the case id column, defaults to XESFormatter.Parameters.CASE_ID
        :type case_id: str, optional
        :param activity_key: name of the activity column, defaults to XESFormatter.Parameters.ACTIVITY
        :type activity_key: str, optional
        :param timestamp_key: name of the timestamp column, defaults to XESFormatter.Parameters.TIMESTAMP
        :type timestamp_key: str, optional
        :param lifecycle_type: name of the event lifecycle column, defaults to XESFormatter.Parameters.TYPE
        :type lifecycle_type: str, optional
        :param timestamp_format: timestamp format, defaults to XESFormatter.Parameters.TIMESTAMP_FORMAT
        :type timestamp_format: str, optional

        Returns
        -------

        :return: Raw event data object
        :rtype: RawEventData

        Raises
        ---------
        ValueError: If the specified event log data is not in dataframe format
        """                
        parameters = {}
        parameters[Constants.CASE_ID_KEY]=case_id
        parameters[Constants.ACTIVITY_KEY]=activity_key
        parameters[Constants.TIMESTAMP_KEY]=timestamp_key
        parameters[Constants.TIMESTAMP_FORMAT_KEY] = timestamp_format  
        parameters[Constants.TYPE_KEY]=lifecycle_type     
        parameters[Constants.STARTTIME_COLUMN]=starttime_column
        formatter = CSVFormatter(parameters)
        extracted_log = formatter.extract_from_dataframe(dataframe)
        return extracted_log

  
def import_mxml(eventlog,case_id: str=MXMLFormatter.Parameters.CASE_ID, activity_key: str=MXMLFormatter.Parameters.ACTIVITY, timestamp_key: str=MXMLFormatter.Parameters.TIMESTAMP, lifecycle_type: str= MXMLFormatter.Parameters.TYPE,timestamp_format: str=MXMLFormatter.Parameters.TIMESTAMP_FORMAT) ->RawEventData:
        """
        Parse MXML file into event log

        Parameters
        -----------
        :param eventlog: XES event log file
        :type eventlog: Path to the file
        :param case_id: name of the case id column, defaults to XESFormatter.Parameters.CASE_ID
        :type case_id: str, optional
        :param activity_key: name of the activity column, defaults to XESFormatter.Parameters.ACTIVITY
        :type activity_key: str, optional
        :param timestamp_key: name of the timestamp column, defaults to XESFormatter.Parameters.TIMESTAMP
        :type timestamp_key: str, optional
        :param lifecycle_type: name of the event lifecycle column, defaults to XESFormatter.Parameters.TYPE
        :type lifecycle_type: str, optional
        :param timestamp_format: timestamp format, defaults to XESFormatter.Parameters.TIMESTAMP_FORMAT
        :type timestamp_format: str, optional

        Returns
        -------

        :return: Raw event data object
        :rtype: RawEventData

        Raises
        ---------
        ValueError: If the specified event log data is not in MXML format
        """        
        parameters = {}
        parameters[Constants.CASE_ID_KEY]=case_id
        parameters[Constants.ACTIVITY_KEY]=activity_key
        parameters[Constants.TIMESTAMP_KEY]=timestamp_key
        parameters[Constants.TIMESTAMP_FORMAT_KEY] = timestamp_format          
        parameters[Constants.TYPE_KEY] = lifecycle_type 
        formatter = MXMLFormatter(parameters)
        dataframe = formatter.extract_data(eventlog)
        return dataframe
    

# def _format_log(dataframe: RawEventData) ->  DataFrame:     
#         event_log = dataframe.getData()   
#         formatted_log = pm4py.format_dataframe(event_log, case_id=dataframe.mandatory_properties[Constants.CASE_ID_KEY], activity_key=dataframe.mandatory_properties[Constants.ACTIVITY_KEY], timestamp_key=dataframe.mandatory_properties[Constants.TIMESTAMP_KEY])                                     
#         #return formatted_log     
#         return event_log


def discover_heuristics_net(dataframe: RawEventData,lifecycleTypes = None) -> HeuristicsNet:        
        """
        Apply heuristic mining algorithm on the RawEventData event log object to discover heuristic net

        :param dataframe: event log
        :type dataframe: RawEventData
        :param lifecycleTypes: lifecycle event types to filter, defaults to None
        :type lifecycleTypes: List, optional
        :return: heuristic net
        :rtype: HeuristicsNet
        """
        event_log = dataframe     
        if (lifecycleTypes is not None) or (Constants.TYPE_KEY in dataframe.getMandatoryProperties()):
                event_log= dataframe.filterLifecycleEvents(lifecycleTypes)
                                   
        formatted_log = event_log.getLog()

        #log = log_converter.apply(formatted_log)
        map =  pm4py.discover_heuristics_net(formatted_log)
        return map

      
   
def view_heuristics_net(map: HeuristicsNet):
        """
        Create view of the heuristic net

        :param map: Heuristic net
        :type map: HeuristicsNet
        """        
        pm4py.view_heuristics_net(map)

def discover_dfg(dataframe: RawEventData,lifecycleTypes = None):        
        """
        Apply dfg mining algorithm on the RawEventData event log object to discover heuristic net

        :param dataframe: event log
        :type dataframe: RawEventData
        :param lifecycleTypes: lifecycle event types to filter, defaults to None
        :type lifecycleTypes: List, optional
        :return: dfg
        :rtype: 
        """
        event_log = dataframe     
        if (lifecycleTypes is not None) or (Constants.TYPE_KEY in dataframe.getMandatoryProperties()):
                event_log= dataframe.filterLifecycleEvents(lifecycleTypes)
                                   
        formatted_log = event_log.getLog()
        dfg =  dfg_discovery.apply(formatted_log, variant=dfg_discovery.Variants.FREQUENCY)
        return dfg, formatted_log

def view_dfg(dfg: dict, formatted_log):
        """
        Create view of the dfg
        """
        gviz = dfg_visualization.apply(dfg, log=formatted_log, variant=dfg_visualization.Variants.FREQUENCY)
        dfg_visualization.view(gviz)

    
def discover_bpmn_model( dataframe: RawEventData) -> BPMN:    
        """
        Performs process mining on the event log data to discover bpmn model

        :param dataframe: event log
        :type dataframe: RawEventData
        :return: BPMN
        :rtype: BPMN
        """            
        process_tree = discover_process_tree(dataframe)
        bpmn_model = pm4py.convert_to_bpmn(process_tree)
        return bpmn_model

  
def view_bpmn_model(bpmn_model: BPMN):
        """
        Create a view of the BPMN model
        :param bpmn_model: BPMN
        :type bpmn_model: BPMN
        """        
        pm4py.view_bpmn(bpmn_model)

    
def discover_process_tree(dataframe: RawEventData,lifecycleTypes = None) ->ProcessTree:
        """
        Perform process mining on the event log to discover process tree

        :param dataframe: _description_
        :type dataframe: RawEventData
        :param lifecycleTypes: _description_, defaults to None
        :type lifecycleTypes: _type_, optional
        :return: _description_
        :rtype: ProcessTree
        """        
        
        event_log = dataframe     
        if (lifecycleTypes is not None) or (Constants.TYPE_KEY in dataframe.getMandatoryProperties()):
                event_log= dataframe.filterLifecycleEvents(lifecycleTypes)
   
        formatted_log = event_log.getLog()

        #log = log_converter.apply(formatted_log)   
        
        process_tree = pm4py.discover_process_tree_inductive(formatted_log)
        return process_tree
    
   
def view_process_tree(process_tree: ProcessTree):
        """
        Create process tree view
        :param process_tree: prpocess tree
        :type process_tree: ProcessTree
        """        
        pm4py.view_process_tree(process_tree)


   
def discover_process_map( dataframe: RawEventData,lifecycleTypes = None) -> Tuple[dict,dict,dict]:
        """
        Discover process map

        :param dataframe: event log
        :type dataframe: RawEventData
        :param lifecycleTypes: event lifecycle types to filter, defaults to None
        :type lifecycleTypes: List, optional
        :return: process map
        :rtype: Tuple[dict,dict,dict]
        """        
        event_log = dataframe     
        if (lifecycleTypes is not None) or (Constants.TYPE_KEY in dataframe.getMandatoryProperties()):
                event_log= dataframe.filterLifecycleEvents(lifecycleTypes)
          
        formatted_log = event_log.getLog()

        #log = log_converter.apply(formatted_log)
        
        dfg, start_activities, end_activities = pm4py.discover_dfg(formatted_log)
        return dfg, start_activities, end_activities
        
    
def view_process_map(dfg, start_activities, end_activities):
        """
        Create a view of process map
        
        :param dfg: dfg
        :type dfg: DFG
        :param start_activities: list of start activities
        :type start_activities: List
        :param end_activities: list of end activities
        :type end_activities: List
        """        
        pm4py.view_dfg(dfg, start_activities, end_activities)
          

def filter_start_activities(dataframe: RawEventData, activities, retain=True):
        """
        Filter cases having a start activity in the provided list

        :param activities: collection of start activities
        :type activities: List
        :param retain: if True, we retain the traces containing the given start activities, if false, we drop the traces
        :type retain: bool, optional
        :return: filtered dataframe
        :rtype: Union[EventLog, pd.DataFrame]
        """        
        filtered = pm4py.filter_start_activities(dataframe.getLog(), activities,retain)
        return filtered
    
def filter_end_activities(dataframe: RawEventData, activities, retain=True):
        """
        Filter cases having an end activity in the provided list

        :param activities: collection of end activities
        :type activities: List
        :param retain: if True, we retain the traces containing the given end activities, if false, we drop the traces
        :type retain: bool, optional
        :return: filtered dataframe
        :rtype: Union[EventLog, pd.DataFrame]
        """        
        filtered = pm4py.filter_end_activities(dataframe.getLog(), activities,retain)
        return filtered
    
def getStartActivities(dataframe: RawEventData):
        """
        Returns the start activities from a log object

        :return: Dictionary of start activities along with their count
        :rtype: dict
        """        
        return pm4py.get_start_activities(dataframe.getLog())
    
def getEndActivities(dataframe: RawEventData):       
        """
        Returns the end activities from a log object

        :return: Dictionary of end activities along with their count
        :rtype: dict
        """      
        return pm4py.get_end_activities(dataframe.getLog())

def getDataProcessRepresentation(dataframe: RawEventData):
        """
        The purpose of this function is to take a raw event log as input and output a dictionary representation of the process model discovered when mining this event log.
        :param dataframe: A pandas dataframe containing the raw event log data.
        :type dataframe: RawEventData
        :return: A dictionary representing the process model, where each key is a tuple representing a transition between two activities, and the value is the strength of that transition as determined by the frequency with which it occurs in the event log.
        :rtype: dict
        """        
        dfg,event_log = discover_dfg(dataframe)
        processRepresentation = getModelProcessRepresentation(dfg)
        return processRepresentation

def getModelProcessRepresentation(model):
        def _getPairsProcess(dfgModel):    
                result_dict = {}
                for pair, strength in dfgModel.items():
                        result_dict[pair] = strength
                return result_dict        
        processRepresentation = _getPairsProcess(model)
        return processRepresentation



# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
from abc import abstractmethod
from typing import Optional

from ...causal_process_discovery.algorithms.base_causal_alg import CausalResultInfo
from ...causal_process_discovery.causal_constants import DEFAULT_VARIANT, Algorithm
from ...process_data.raw_event_data import RawEventData


class BaseAnchor:
    """
    Parent class for all causal discovery modalities. Please extend this class to add more modalities. Different modalities are described here <>
    """    
 
    @abstractmethod
    def apply(self, dataObject: RawEventData,variant: Optional[Algorithm] = DEFAULT_VARIANT) -> CausalResultInfo:
        """
        Apply the chosen algorithm variant with the current modality on the process data object.

        :param dataObject:  event log
        :type dataObject: RawEventData
        :param variant: the chosen algorithm variant to apply, defaults to LINGAM
        :type variant: Optional[Algorithm], optional
        :return: causal discovery result
        :rtype: CausalResultInfo
        """    
        return NotImplementedError
        


# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
from enum import Enum
from typing import Optional

import pandas as pd

from sax.core.process_data.data import BaseProcessDataObject
from sax.core.process_data.raw_event_data import RawEventData
from sax.core.process_data.tabular_data import TabularEventData
from .constants import Constants


def unroll(value):
    if isinstance(value, Enum):
        return value.value
    return value


def get_param_value(p, parameters, default):
    if parameters is None:
        return unroll(default)
    unrolled_parameters = {}
    for p0 in parameters:
        unrolled_parameters[unroll(p0)] = parameters[p0]
    if p in parameters:
        val = parameters[p]
        return unroll(val)
    up = unroll(p)
    if up in unrolled_parameters:
        val = unrolled_parameters[up]
        return unroll(val)
    return unroll(default)

#Not sure need this
def get_properties(dataObject:BaseProcessDataObject, activity_key: str = "concept:name", timestamp_key: str = "time:timestamp", case_id_key: str = "case:concept:name", resource_key: str = "org:resource", group_key: Optional[str] = None,timestamp_format: Optional[str] = None, **kwargs):
    """
    Gets the properties from an event data object signifying the column names of the mandatory attributes, date formatter string, and additional attributes

    :param dataObject: data object
    :param activity_key: attribute to be used for the activity
    :param timestamp_key: attribute to be used for the timestamp
    :param case_id_key: attribute to be used as case identifier
    :param resource_key: (if provided) attribute to be used as resource
    :param group_key: (if provided) attribute to be used as group identifier
    :rtype: ``Dict``
    """
  

    if type(dataObject) not in [RawEventData, TabularEventData]: return {}

    from copy import copy
    parameters = copy(dataObject.properties) if hasattr(dataObject, 'properties') else copy(dataObject.attrs) if hasattr(dataObject,
                                                                                                    'attrs') else {}

    if activity_key is not None:
        parameters[Constants.ACTIVITY_KEY] = activity_key        

    if timestamp_key is not None:
        parameters[Constants.TIMESTAMP_KEY] = timestamp_key

    if case_id_key is not None:
        parameters[Constants.CASE_ID_KEY] = case_id_key

    if resource_key is not None:
        parameters[Constants.RESOURCE_KEY] = resource_key

    if group_key is not None:
        parameters[Constants.GROUP_KEY] = group_key
    
    if timestamp_format is not None:
        parameters[Constants.TIMESTAMP_FORMAT] = timestamp_format

    for k, v in kwargs.items():
        parameters[k] = v

    return parameters

def convert_timestamp_columns_in_df(df, timest_format=None, timest_columns=None):    
    for col in df.columns:
        if timest_columns is None or col in timest_columns:
            if "obj" in str(df[col].dtype) or "str" in str(df[col].dtype):
                try:
                    if timest_format is None:
                        # makes operations faster if non-ISO8601 but anyhow regular dates are provided
                        df[col] = pd.to_datetime(df[col], utc=True)
                    else:
                        df[col] = pd.to_datetime(df[col], format=timest_format)
                        df[col] = df[col].dt.tz_localize('UTC')
                except Exception as e:                    
                    try:
                        df[col] = pd.to_datetime(df[col], utc=True)
                    except:
                        pass
    return df

# Calculate start time of each trace and add it as separate column to each row event
def add_start_time(df,timestamp_column_name,id_column_name, start_column_name):    
    #TODO: do we want to impose 'start' lifecycle transition, or just go with the first timestamp?
    #start_events = df[df[lifecycle_event_column_name].str.lower() == start_timestamp_value]
    #if start_events.empty:
        #raise ValueError("The event log activities has no start lifecycle events, cannot proceed to calculate trace start times")

    # Step 2: Group the remaining data by 'id' and sort by 'timestamp'
    sorted_data = df.sort_values(by=[id_column_name, timestamp_column_name])

    # Step 3: For each group, get the first timestamp and create a new DataFrame
    start_times = sorted_data.groupby(id_column_name).first()[timestamp_column_name].reset_index()
    start_times.columns = [id_column_name, start_column_name]

    # Step 4: Merge the original DataFrame with the start_times DataFrame
    result_df = pd.merge(df, start_times, on=id_column_name, how='left')

    # Display the result
    return result_df



# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
class ValidationException(Exception):
    """Exception risen then there is validation exception on the passed parameters
    """
    def __init__(self, message, details):
        self.message = message
        self.details = details
        super().__init__(self.message)


# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
from langchain_community.document_loaders.pdf import PyPDFLoader
from langchain_community.document_loaders.csv_loader import CSVLoader
from langchain_community.document_loaders.directory import DirectoryLoader
from langchain_community.document_loaders.text import TextLoader
import os
from langchain_text_splitters import RecursiveCharacterTextSplitter
from langchain.chains.retrieval_qa.base import RetrievalQA
from langchain.vectorstores.chroma import Chroma
from sax.core.synthesis.exceptions.validation import ValidationException
from sax.core.synthesis.llms.base_llm import BaseLLM
from sax.core.synthesis.rag.base_retrieval import BaseRetriever


CHROMA_PATH = "db/"

class DocumentRetrieverLLM(BaseRetriever):    
    """A document retrieval imeplementation based on Chroma vector store

    """

    def __init__(self,modelClass: BaseLLM,docLocation,filterFiles=None, chunk_size=None, chunk_overlap=None,retrieved_results=None, dbPath=None):
        self._model = modelClass
        self._retriever = self._indexDocuments(modelClass,docLocation,filterFiles,chunk_size,chunk_overlap,retrieved_results,dbPath)

    def _getDocuments(self,docPath,glob = None):
        """Load all the documents in the given path with the given type and return the loaded docs as array
    
        :param docPath: path to the document/folder
        :type docPath: str
        :param glob: type of documents to embed (txt,pdf,csv), defaults to None
        :type glob: str, optional
        :raises ValidationException: unknown document types in the specified path
        :return: array of loaded documents
        :rtype: array
        """
        documents = []
        if docPath.endswith('.csv'):
            loader =  CSVLoader(docPath)
            documents = loader.load()
        elif docPath.endswith('.pdf'):
            loader =  PyPDFLoader(docPath)
            documents = loader.load()
        elif docPath.endswith('.txt'):             
            loader = TextLoader(docPath)
            documents = loader.load()
        elif os.path.isdir(docPath) and glob:
            if glob == "pdf":
                for file in os.listdir(docPath):
                    if file.endswith('.pdf'):
                        pdf_path = os.path.join(docPath, file)
                        loader = PyPDFLoader(pdf_path)
                        documents.extend(loader.load())
            elif glob == "csv":
               for file in os.listdir(docPath):
                    if file.endswith('.csv'):
                        csv_path = os.path.join(docPath, file)
                        loader = CSVLoader(csv_path)
                        documents.extend(loader.load())                
            elif glob =="txt": 
               for file in os.listdir(docPath):
                    if file.endswith('.txt'):
                        txt_path = os.path.join(docPath, file)
                        loader = TextLoader(txt_path)                        
                        documents.extend(loader.load())                         
        else: raise ValidationException("ValidationError: no appropriate loader for doc: ",docPath)
        return documents


    def _indexDocuments(self, model: BaseLLM,docLocation,filterFiles=None,chunk_size=None,chunk_overlap=None,retrieved_results=None,dbPath=None):
        """Load the documents in the given location , index and store in the Chroma vector store for later contextual retrieval.
        
        :param model: the wrapper for the chosen model provider for embedding model
        :type model: BaseLLM
        :param docLocation: path to the document/documents folder
        :type docLocation: str
        :param filterFiles: extension type of the files to load if any
        :type filterFiles: str, optional
        :param chunk_size: the size of the chunk in text splitter, defaults to None
        :type chunk_size: int, optional
        :param chunk_overlap: the overlap of chunks in text splitter, defaults to None
        :type chunk_overlap: int, optional
        :param retrieved_results: the number of results to retrieve, defaults to None
        :type retrieved_results: int, optional
        :param dbPath: path to the Chroma db, defaults to None
        :type dbPath: str, optional
        :return: retriever wrapper to use for later retreivals agains the specified docs
        :rtype: BaseRetriever
        """
        # load the doc/docs
        pages = self._getDocuments(docLocation,filterFiles)        

        # split the doc into smaller chunks i.e. chunk_size=500
        chunk_size_to_pass = chunk_size if chunk_size is not None else 500
        chunk_overlap_to_pass = chunk_overlap if chunk_overlap is not None else 50
        text_splitter = RecursiveCharacterTextSplitter(chunk_size=chunk_size_to_pass, chunk_overlap=chunk_overlap_to_pass)
        chunks = text_splitter.split_documents(pages)

        # get embedding model
        embeddings = model.getEmbeddingModel()

        # embed the chunks as vectors and load them into the database
        dbpath_to_pass = dbPath if dbPath is not None else CHROMA_PATH
        db= Chroma.from_documents(chunks, embeddings, persist_directory=dbpath_to_pass)
        # expose this index in a retriever interface
        kwargs_to_pass= retrieved_results if retrieved_results is not None else 10
        retriever = db.as_retriever(search_type="similarity", search_kwargs={"k":kwargs_to_pass})

        # create a chain to answer questions 
        return retriever
    

    # ----- Retrieval and Generation Process -----

    def get_retriever(self):
        """Return the embedded retriever stored in this wrapper"""
        return self._retriever
 
   


# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
from abc import abstractmethod

class BaseRetriever: 
    """
    Base class for all document retriever wrappers   
    """ 
    @abstractmethod                
    def get_retriever(self):
        """Get the embedded retriever
        """
        pass




# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
from threading import Lock

class State:
    state_lock = Lock()

    def __init__(self):
       self.data= None

    def get_state_copy(self):
        # Acquire the lock before reading app.state
        self.state_lock.acquire()
        try:
            state_copy = self.data.copy() if self.data else None
        finally:
            # Release the lock after reading app.state
            self.state_lock.release()
        return state_copy

    def update_state(self,data):
        self.state_lock.acquire()
        try:
            self.data=data
        finally:
            self.state_lock.release()        

    def clear_state(self):
        self.state_lock.acquire()
        try:
            self.data=None
        finally:
            self.state_lock.release() 



# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
import xml.etree.ElementTree as Xet

import pandas as pd

from sax.core.process_data.data import BaseProcessDataObject
from sax.core.process_data.raw_event_data import RawEventData
from sax.core.utils import helper_utils
from sax.core.utils.constants import Constants, ConstantsMeta
from .base_formatter import BaseFormatter


class MXMLConstants(metaclass=ConstantsMeta):
        """
        Constants representing standart metafields in the MXML format
        
        - PROCESS_INSTANCE: designatation of a the process entry
        - TRACE_INSTANCE: designation of separate trace within the process
        - ID_ATTRIBUTE: designation of the attribute representing trace id
        - EVENT_TYPE: designation of event lifecycle element
        - ACTIVITY: designation of activity element
        - RESOURCE: designation of resource element
        - ATTRIBUTE: designation of attribute element
        - ATTRIBUTE_NAME: designation of attribute name field within attribute element
        """    
        PROCESS_INSTANCE = ".//ProcessInstance"
        TRACE_INSTANCE = ".//AuditTrailEntry"
        ID_ATTRIBUTE = "id"
        EVENT_TYPE = "EventType"
        TIMESTAMP = "Timestamp"
        ACTIVITY = "WorkflowModelElement"
        RESOURCE = "originator"
        ATTRIBUTE = "attribute"
        ATTRIBUTE_NAME = "name"
        DATA = "Data"
        ELEMENT_ID="elementId"
        PROCESS_ID="processId"

    

class MXMLFormatter(BaseFormatter):
    '''
    Formatter class for MXML event log files transforming the file to standard event log dataframe representation
    '''             
    class Parameters(metaclass=ConstantsMeta):
        """
        Constants representing various default values for mandatory attributes values for MXML-based event log

        Attributes
        ----------
        ACTIVITY : str
            Default value for activity key.
        CASE_ID : str
            Default value for case id key.
        TYPE : str
            Default value for lifecycle event key.
        TIMESTAMP : str
            Default value for timestamp key.
        TIMESTAMP_FORMAT : str
            Default value for timestamp format key.
        RESOURCE_ID : str
            Default value for resource id key.
        """
        ACTIVITY = "concept:name"
        CASE_ID = "case:concept:name"
        TYPE = "lifecycle:transition"
        TIMESTAMP = "time:timestamp"
        TIMESTAMP_FORMAT = "%Y-%m-%d %H:%M:%S.%f"
        RESOURCE_ID = "org:resource"   
   
    def __init__(self,  parameters=None):
        """
        Initializes the MXML formatter with relevant information about the MXML parameters and how to extract relevant information from MXML

        Args:
        :param parameters: Mapping between MXML based attributes and resulting event log attributes
        :type parameters: dict

        Raises:
        :raises ValueError: If any of the mandatory parameters are missing or invalid.
        """
        super().__init__(parameters=parameters)
        self.parameters[Constants.ACTIVITY_KEY]= helper_utils.get_param_value(Constants.ACTIVITY_KEY, parameters, MXMLFormatter.Parameters.ACTIVITY)
        self.parameters[Constants.CASE_ID_KEY] = helper_utils.get_param_value(Constants.CASE_ID_KEY, parameters, MXMLFormatter.Parameters.CASE_ID)
        self.parameters[Constants.TIMESTAMP_KEY] = helper_utils.get_param_value(Constants.TIMESTAMP_KEY, parameters, MXMLFormatter.Parameters.TIMESTAMP)
        self.parameters[Constants.TYPE_KEY] = helper_utils.get_param_value(Constants.TYPE_KEY, parameters, MXMLFormatter.Parameters.TYPE)
        self.parameters[Constants.TIMESTAMP_FORMAT_KEY] = helper_utils.get_param_value(Constants.TIMESTAMP_FORMAT_KEY, parameters, MXMLFormatter.Parameters.TIMESTAMP_FORMAT)        
        self.parameters[Constants.RESOURCE_KEY] = helper_utils.get_param_value(Constants.RESOURCE_KEY, parameters, MXMLFormatter.Parameters.RESOURCE_ID)        

    def extract_data(self,event_log_data) -> BaseProcessDataObject:
        """
        Extract tabular data from the provided MXML event log file, instantiate raw data object containing the tabular data.

        Parameters
        -----------
        :param event_log_data: File representing the event log data, in MXML format
        :type event_log_data: file

        Returns
        -------
        :return: A RawEventData representing the event log in a raw event format
        :rtype: RawEventData

        Raises
        ------
        :raises ValueError: If the event log data is not in MXML format.
        """ 
        df = self._parse_xml(event_log_data)        
        original_data = pd.DataFrame(df)       
        event_log = helper_utils.convert_timestamp_columns_in_df(original_data, timest_format=self.parameters[Constants.TIMESTAMP_FORMAT_KEY], timest_columns=self.parameters[Constants.TIMESTAMP_KEY])
        mandatory_properties,optional_properties = self._getProperties(event_log, self.parameters)
        event_log = helper_utils.add_start_time(event_log, timestamp_column_name=mandatory_properties[Constants.TIMESTAMP_KEY], id_column_name=mandatory_properties[Constants.CASE_ID_KEY], start_column_name=Constants.STARTTIME_COLUMN)
        mandatory_properties[Constants.STARTTIME_COLUMN]=Constants.STARTTIME_COLUMN     
        return RawEventData(event_log,mandatory_properties,optional_properties)
    
        
    

    def  _parse_xml(self,event_log_data) -> pd.DataFrame:
        """
        Parse the XML event log data into a Pandas DataFrame.

        Parameters
        -----------
        :param event_log_data: File representing the event log data, in MXML format
        :type event_log_data: file

        Returns
        --------
        :return: A Pandas DataFrame representation of the event log data.
        :rtype: pd.DataFrame

        Raises
        ------
        :raises ValueError: If the event log data is not in MXML format.
        """
        tree = Xet.parse(event_log_data)
        root = tree.getroot()
        rows = []       
        for process_instance in root.findall(MXMLConstants.PROCESS_INSTANCE):
            process_instance_id = process_instance.get(MXMLConstants.ID_ATTRIBUTE)
            for audit_entry in process_instance.findall(MXMLConstants.TRACE_INSTANCE):
                data_dict = {}
                resource_element = None
                for child_element in audit_entry:
                    tag_name_lower = child_element.tag.lower()
                    if tag_name_lower == MXMLConstants.ACTIVITY.lower():
                        activity = child_element.text
                    elif tag_name_lower == MXMLConstants.EVENT_TYPE.lower():
                        event_type = child_element.text
                    elif tag_name_lower == MXMLConstants.TIMESTAMP.lower():
                        timestamp = child_element.text
                    elif tag_name_lower == MXMLConstants.RESOURCE.lower():
                        resource_element = child_element
                    elif tag_name_lower == MXMLConstants.DATA.lower():
                        data_element = child_element
                        attribute_elements = [element for element in data_element if element.tag.lower() == MXMLConstants.ATTRIBUTE]
                        for attr in attribute_elements:
                            attr_name = attr.get(MXMLConstants.ATTRIBUTE_NAME)
                            #use only those attribute which do not appear in the mandatory attributes list
                            if (attr_name.lower() != Constants.CASE_ID_KEY.lower() and attr_name.lower() != Constants.ACTIVITY_KEY.lower() and attr_name.lower() != Constants.TYPE_KEY.lower() and attr_name.lower() != Constants.TIMESTAMP_KEY.lower() and attr_name.lower() != Constants.RESOURCE_KEY.lower() and attr_name.lower() != MXMLConstants.PROCESS_ID.lower() and attr_name.lower() != MXMLConstants.ELEMENT_ID.lower()):
                                attr_value = attr.text
                                new_attr_name = f"Attr_{attr_name}"  # Add a prefix to avoid name collision
                                data_dict[new_attr_name] = attr_value   
                                       
                resource = resource_element.text if resource_element is not None else None

                row = {
                            self.parameters[Constants.CASE_ID_KEY]: process_instance_id,
                            self.parameters[Constants.ACTIVITY_KEY]: activity,
                            self.parameters[Constants.TYPE_KEY]: event_type,
                            self.parameters[Constants.TIMESTAMP_KEY]: timestamp,
                            self.parameters[Constants.RESOURCE_KEY]: resource,
                            **data_dict
                        }
                rows.append(row)
        df = pd.DataFrame(rows)
        return df


# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
from typing import Any, Dict, Tuple, List

import pandas as pd
from pandas import DataFrame

from sax.core.utils.constants import Constants


class BaseProcessDataObject:
    '''
    Data object for process event log data. The data object contains dataframe created from various event log formats 
    holding basic process event log data, such as trace id, activity name, activity timestamp.
    '''
    def __init__(self, data:DataFrame, mandatory_properties:dict,optional_properties: dict):
        """
        Creates data object representing the event log.

        Args:
        :param data: : dataframe representing event log data
        :type data: DataFrame

        :param mandatory_properties: Dictionary representing a mapping between mandatory standard event log attribute names [ "case:concept:name","concept:name","lifecycle:transition","time:timestamp","org:resource"] and name of corresponding columns in the dataframe
        :type mandatory_properties: dict

        :param optional_properties: Dictionary representing a mapping between optional attribute names and names of corresponding columns in the dataframe
        :type optional_properties: dict
        """ 
        if type(data) not in [pd.DataFrame]: raise TypeError("the method can be applied only to a dataframe!")
      
        self.data = data
        self.mandatory_properties = dict(mandatory_properties)      
        self.optional_properties = dict(optional_properties) 
       
    def __str__(self):
        """
        Return string representaion of the data class content
        :return: Dataframe content, mandatory properties and optional properties
        :rtype: string
        """        
        return f"Data: {self.data},\n Mandatory properties: {self.mandatory_properties} \n Optional properties: {self.optional_properties}"
    
    def copy(self):
        """
        Make a deep copy of the data object.

        :return: A copy of the BaseProcessDataObject instance.
        :rtype: BaseProcessDataObject
        """    
        # Make a deep copy of the DataFrame
        data_copy = self.data.copy(deep=True)
        # Make shallow copies of the dictionaries
        mandatory_properties_copy = self.mandatory_properties.copy()
        optional_properties_copy = self.optional_properties.copy()
        # Create a new instance of BaseProcessDataObject with copied data
        return BaseProcessDataObject(data_copy, mandatory_properties_copy, optional_properties_copy)
    
    def getData(self) -> DataFrame:
        """
        Return the dataframe represeting the event log part of the data object

        :return: A dataframe representing the event log 
        :rtype: DataFrame
        """ 
        return pd.DataFrame(self.data)
    
    def getMandatoryProperties(self) -> dict:
        """
        Get mandatory properties of the data object

        :return: Mandatory properties mapping
        :rtype: dict
        """        
        return self.mandatory_properties
    
    def getOptionalProperties(self) -> dict:
        """
        Get optional properties of the data object

        :return: Optional properties mapping
        :rtype: dict
        """   
        return self.optional_properties

    def getCaseIdColumnName(self) -> str:
        """
        Get the name of the column representing case id

        :return: CaseId column name
        :rtype: str
        """        
        return self.mandatory_properties[Constants.CASE_ID_KEY]
    

    def _extract_properties(self,df, mandatory_properties) -> Tuple[dict, dict]:
        """
        Extract the optional and mandatory properties from the dataframe columns.

        :param df: The input dataframe.
        :type df: pd.DataFrame
        :param mandatory_properties: The mandatory properties dictionary.
        :type mandatory_properties: dict
        :return: The extracted mandatory and optional properties dictionaries.
        :rtype: Tuple[dict, dict]
        """
        mandatory_columns_names = [col for col in df.columns if "__" not in col]
        optional_column_names = [col for col in df.columns if "__" in col]
        #create dictionaries from the column names
        Id_column_name = mandatory_properties[Constants.CASE_ID_KEY]
      
        mandatory_dict = {col: col for col in mandatory_columns_names if col is not mandatory_properties[Constants.CASE_ID_KEY]} 
        mandatory_dict[Constants.CASE_ID_KEY]=Id_column_name

        optional_dict = {col: col for col in optional_column_names}
        return mandatory_dict, optional_dict
    

    def getActivitiesForTrace(self,pid:str) -> list[str]:
        """
        Return a list of all activities for the provided case id

        :param pid: process id
        :type pid: str
        :return: List of all activity names
        :rtype: list[str]
        """            
        return NotImplementedError
    

    def getTrace(self,pid:str) -> Dict[str, Any]:
        """
        Return all the information regarding the given trace in a dictionary, where the keys are the activity names and the values are the activity payloads

        :return: A dictionary representing a single trace
        :rtype: Dict[str, List[str]]
        """                    
        return NotImplementedError  
            

    def getVariant(self, activities: List[str]) -> 'BaseProcessDataObject':
        """
        Provided a list of activity names, return in a dataframe all traces instances which belong to this variant

        :param activities: List of activities names
        :type: List
        :return: Dataframe with all trace instances containing specified activities
        :rtype: BaseProcessDataObject
        """        
        return NotImplementedError  
    
    def getVariants(self) -> dict:
        """
        Return a dictionary mapping of all possible process variants in the event log, to a number of traces per each variant
        
        :return:  Mapping of all possible process variants in the event log, to a number of traces per each variant
        :rtype: dict
        """        
        return NotImplementedError  
    
    def getLength(self) -> int:
        """
        Return the number of traces in the event log

        :return: Length of the event log
        :rtype: int
        """        
        
        return len(self.data)
        
    
        
    

   
  


# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
import pm4py

from sax.core.process_data.raw_event_data import RawEventData
from sax.core.utils import helper_utils
from sax.core.utils.constants import Constants, ConstantsMeta
from .base_formatter import BaseFormatter


class XESFormatter(BaseFormatter):
    '''
    Formatter class for XES event log files transforming the file to standard event log dataframe representation
    '''      

    class Parameters(metaclass=ConstantsMeta):
        """
        Constants representing various default values for mandatory attributes values for XES-based event log

        Attributes
        ----------
        ACTIVITY : str
            Default value for activity key.
        CASE_ID : str
            Default value for case id key.
        TYPE : str
            Default value for lifecycle event key.
        TIMESTAMP : str
            Default value for timestamp key.
        TIMESTAMP_FORMAT : str
            Default value for timestamp format key.
        RESOURCE_ID : str
            Default value for resource id key.
        """
        ACTIVITY = "concept:name"
        CASE_ID = "case:concept:name"
        TYPE = "lifecycle:transition"
        TIMESTAMP = "time:timestamp"
        TIMESTAMP_FORMAT = "%Y-%m-%d %H:%M:%S.%f"
        RESOURCE_ID = "org:resource" 
  
    def __init__(self,  parameters=None):
        """
        Initializes the XES formatter with relevant information about the MXML parameters and how to extract relevant information from XES

        Args:
        :param parameters: Mapping between XES based attributes and resulting event log attributes
        :type parameters: dict

        Raises:
        :raises ValueError: If any of the mandatory parameters are missing or invalid.
        """
        super().__init__(parameters=parameters)
        self.parameters[Constants.ACTIVITY_KEY]= helper_utils.get_param_value(Constants.ACTIVITY_KEY, parameters, XESFormatter.Parameters.ACTIVITY)
        self.parameters[Constants.CASE_ID_KEY] = helper_utils.get_param_value(Constants.CASE_ID_KEY, parameters, XESFormatter.Parameters.CASE_ID)
        self.parameters[Constants.TIMESTAMP_KEY] = helper_utils.get_param_value(Constants.TIMESTAMP_KEY, parameters, XESFormatter.Parameters.TIMESTAMP)
        self.parameters[Constants.TYPE_KEY] = helper_utils.get_param_value(Constants.TYPE_KEY, parameters, XESFormatter.Parameters.TYPE)
        self.parameters[Constants.TIMESTAMP_FORMAT_KEY] = helper_utils.get_param_value(Constants.TIMESTAMP_FORMAT_KEY, parameters, XESFormatter.Parameters.TIMESTAMP_FORMAT)        

    def extract_data(self,event_log_data) -> RawEventData:
        """
        Extract tabular data from the provided XES event log file, instantiate raw data object containing the tabular data.

        Args:
        :param event_log_data: File representing the event log data, in XES format
        :type event_log_data: file

        Returns:
        :return: A RawEventData representing the event log in a raw event format
        :rtype: RawEventData

        Raises:
        :raises ValueError: If the event log data is not in MXML format.
        """      
        event_log = pm4py.read_xes(event_log_data)        
        dataframe = pm4py.convert_to_dataframe(event_log)   
        event_log = helper_utils.convert_timestamp_columns_in_df(dataframe, timest_format=self.parameters[Constants.TIMESTAMP_FORMAT_KEY], timest_columns=self.parameters[Constants.TIMESTAMP_KEY])
        mandatory_properties,optional_properties = self._getProperties(dataframe, self.parameters) 
        event_log = helper_utils.add_start_time(event_log, timestamp_column_name=mandatory_properties[Constants.TIMESTAMP_KEY], id_column_name=mandatory_properties[Constants.CASE_ID_KEY], start_column_name=Constants.STARTTIME_COLUMN)
        mandatory_properties[Constants.STARTTIME_COLUMN]=Constants.STARTTIME_COLUMN          
        return RawEventData(event_log,mandatory_properties,optional_properties)
       


# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
class XAILibrary:
    """
    Class for generating explanations for process behaviors in event logs using various XAI algorithms.
    """
    def __init__(self, event_log_data):
        """
        Initializes the XAILibrary with event log data.

        Parameters:
        event_log_data (list): A list of event log data where each element represents an event.
                               The data can be in any appropriate format for your use case.

        Raises:
        NotImplementedError: If the user attempts to call this method directly, this exception will be raised.
        """
        self.event_log_data = event_log_data

    def explain_process_behaviors(self):
        """
        Generates explanations for the process behaviors in the event log data.

        Returns:
            dict: A dictionary representing explanations for the process behaviors, such as feature importance or rule-based explanations.

        Raises:
            NotImplementedError: If the user attempts to call this method directly, this exception will be raised.
        """
        # Perform XAI algorithm here and generate process behavior explanations

        behavior_explanations = {
            "activity1": "This activity is important due to its high occurrence rate.",
            "activity2": "This activity has a significant impact on the overall process duration.",
            "activity3": "This activity often leads to delays in subsequent activities.",
            "activity4": "This activity is a critical step in the process."
        }
        
        return behavior_explanations


# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
from typing import Optional

from pandas import DataFrame

import lingam
from sax.core.causal_process_discovery.algorithms.base_causal_alg import BaseCausalAlgorithm, CausalDataException, \
    CausalResultInfo
from sax.core.causal_process_discovery.prior_knowledge import PriorKnowledge


class LingamImpl(BaseCausalAlgorithm):
    """DirectLINGAM causal discovery algorithm wrapper for process execution causal discovery. Make use of `DirectLINGAM algorithm <https://lingam.readthedocs.io/en/latest/reference/direct_lingam.html>`_

    :param BaseCausalAlgorithm: base type
    :type BaseCausalAlgorithm: BaseCausalAlgorithm
    """    
    def __init__(self,   data: DataFrame, prior_knowledge: Optional[PriorKnowledge]=None):      
        super().__init__(data,prior_knowledge)

    def sanity_check(self): 
        """
        Sanity check for LINGAM, validates that the data on which it is run does not contain NaN values.

        :raises CausalDataException: throws exception in case data contains NaN values.
        """               
        if self.data.isna().any().any():
            raise  CausalDataException("Cannot apply Lingam on dataframe with NaN values, if using chain anchor please first filter the data to the desired variant")

    def run(self) ->CausalResultInfo:
        """
        Runs DirectLINGAM on the data the algorithm instance was initiated with, and returns result object containing LINGAM adjaceny matrix.

        :return: CausalResultInfo
        :rtype: CausalResultInfo
        """       
        #TODO implement differently with/without prior knowledge,target variables etc.  
        self.sanity_check()     
        if self.prior_knowledge is not None:
            model = lingam.DirectLiNGAM(prior_knowledge=self.prior_knowledge.getPriorKnowledge())    
        else:
            model = lingam.DirectLiNGAM()    
        model.fit(self.data)
        
        return CausalResultInfo(model.adjacency_matrix_,list(self.data.columns))
    



import sys
from setuptools import setup, find_packages

def parse_requirements(filename):
    try:
        with open(filename, "r", encoding="utf-8") as fh:
            return [line.strip() for line in fh if line.strip() and not line.startswith('#')]
    except Exception as e:
        print(f"Error reading {filename}: {e}")
        return []
    
with open("README.md", "r",encoding="utf-8") as fh:
    long_description = fh.read()

install_requires = parse_requirements("requirements.txt")

setup(
    name="sax4bpm", 
    version="0.1.1", 
    author="Inna Skarbovsky",
    author_email="inna@il.ibm.com",
    description="Open source Python library for deriving explanations about business processes based on process,causal and XAI perspectives",
    long_description=long_description,
    license='GPL 3.0',
    long_description_content_type="text/markdown",
    url="https://github.com/IBM/sax4bpm",
    packages=find_packages(include=['sax', 'sax.*']),    
    python_requires='>=3.9',
    install_requires=install_requires,
    include_package_data=True,
)



# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
from abc import abstractmethod
from typing import List, Optional

from pandas import DataFrame

from sax.core.causal_process_discovery.prior_knowledge import PriorKnowledge
from sax.core.process_data.discovery_result import ResultInfo


class CausalDataException(Exception):
    """
    Exception indicating that the data object containing process data is not suitable for causal discovery in its present form.

    :param Exception: exception indicating the reason why the process data object is not suitable for causal discovery
    :type Exception:
    """    
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

class CausalResultInfo(ResultInfo):
    # Assisted by WCA for GP
    # Latest GenAI contribution: granite-20B-code-instruct-v2 model
    """
    Class containing causal discovery results. Basically this is a wrapper around causal adjacency matrix.
    """
  
    def __init__(self, adjacencyMatrix: List[List[int]], columns: List[str]):
       super().__init__()
       self.adjacencyMatrix = adjacencyMatrix
       self.columns = columns

    def getAdjacencyMatrix(self) -> List[List[int]]:
        """
        Return causal adjacency matrix contained within this object

        :return: causal adjacency matrix
        :rtype: List[List[int]]
        """        
        return self.adjacencyMatrix
    
    def getColumns(self) -> List[str]:
        """
        Return a list with the names of the columns appearing in the adjacency matrix

        :return: List of column names
        :rtype: List[str]
        """        
        return self.columns
    
    #TODO convert the result to string
    def getDiscoveryResult(self) -> str:
        """
        Return a NL description of the causal relationships 

        :return: Textual description of causal relationships
        :rtype: str
        """        
        return self.adjacencyMatrix
    


class BaseCausalAlgorithm(object):
    # Assisted by WCA for GP
    # Latest GenAI contribution: granite-20B-code-instruct-v2 model
    """
    Parent class for all causal discovery algorithms implementations. Please extend this class to add more algorithms.    
    """
  

    def __init__(self, data: DataFrame, prior_knowledge: Optional[PriorKnowledge]=None):
        """
        Initialize an instance of the algorithm for particular data.

        Parameters
        ----------
        data : pandas.DataFrame
            Dataframe containing the event log.
        prior_knowledge : PriorKnowledge, optional
            Prior knowledge object containing the domain knowledge. Defaults to None.
        """    
        self.data = data
        self.prior_knowledge =prior_knowledge

    @abstractmethod    
    def sanity_check(self):
        """
        Sanity check of the data object for causal discovery

        Returns
        -------
        :return: Boolean indicator whether the data object is coherent
        :rtype: bool

        Raises
        --------
        :raises NotImplementedError:       
        """       
        raise NotImplementedError()
   
    @abstractmethod
    def run(self) ->ResultInfo:
        """
        Run the causal discovery algorithm and return the result
        
        Returns
        -------
        :return: A object holding the causal discovery result
        :rtype: ResultInfo

        Raises
        --------
        :raises NotImplementedError:
        
        """     
        raise NotImplementedError()

    


# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
from abc import abstractmethod
from typing import Tuple

from pandas import DataFrame

from sax.core.process_data.raw_event_data import RawEventData


class BaseFormatter:

    '''
    Formatter base class which can transform event log data from a file in different event log formats to standard raw event log representation for further use.
    '''      

    def __init__(self, parameters = None):
        """
        Initializes the formatter.

        Parameters
        ----------
        parameters : dict, optional
            A mapping of mandatory event attribute names to actual names in the log. The mandatory attribute names include: Constants.CASE_ID_KEY, Constants.ACTIVITY_KEY, Constants.TIMESTAMP_KEY, Constants.TIMESTAMP_FORMAT_KEY, Constants.TYPE_KEY, by default None

        Raises
        ------
        TypeError
            If the input parameter is not a dictionary
        """               
        # set the parameters
        if (parameters is None): parameters = dict()
        self.parameters = dict(parameters)
    
    @abstractmethod
    def extract_data(self, event_log_data) -> RawEventData:
        """
        Create RawEventData data object from the provided event log file

        Parameters
        ----------
        event_log_data : The path designation of the event log file.
        type event_log_data : File

        Returns
        -------
        A RawEventData representing the event log in a tabular format.
        type : RawEventData        
        """
        pass


    def _getProperties(self,dataframe:DataFrame, parameters: dict) -> Tuple[dict, dict]:
        """
        Extracts the properties of the dataframe based on the given parameters.

        Parameters
        ----------
        dataframe : pandas.DataFrame
            The dataframe to extract the properties from.
        parameters : dict
            A dictionary containing the parameters to extract. The keys are the keywords to use, and the values are the names of the columns in the dataframe.

        Returns
        -------
        tuple[dict, dict]
            A tuple containing two dictionaries. The first contains the mandatory properties, and the second contains the optional properties.

        Raises
        ------
        TypeError
            If the dataframe is not a pandas.DataFrame.
        ValueError
            If any of the parameters are not present in the dataframe.
        """
        if type(dataframe) not in [DataFrame]: raise TypeError("the method can be applied only to a dataframe!")              
        mandatory_dict={}
        optional_dict={}
        for keyword, column_name in parameters.items():
            if column_name in dataframe.columns:
                mandatory_dict[keyword] = column_name
    
        for column_name in dataframe.columns:
            if column_name not in parameters.values():
                optional_dict[column_name] = column_name
    
        return mandatory_dict,optional_dict

        


# -----------------------------------------------------------------------------
# Copyright contributors to the SAX4BPM project
# -----------------------------------------------------------------------------
import pandas as pd
from pandas import DataFrame
from pm4py.algo.filtering.log.variants import variants_filter
from pm4py.objects.conversion.log import converter as log_converter

from .data import BaseProcessDataObject
from .tabular_data import TabularEventData
from ..utils.constants import Constants, LifecycleTypes


class RawEventData(BaseProcessDataObject):
    
    """
    Raw event representation of event log, where each row represents a single event in a trace (one activity), and the columns
    are the activity name, activity timestamp, and additional activity payload attributes

    """    

    _permutations = None
    def __init__(self,data:DataFrame, mandatory_properties:dict,optional_properties:dict):           
        """
        Initializes a SAX raw event object.

        Parameters
        ----------
        data : pandas.DataFrame
            The input event log
        mandatory_properties : dict
            A dictionary containing the constants of the mandatory properties as keys and the names of corresponding columns in the dataframe
        optional_properties : dict
            A dictionary containing the names of the optional properties as keys and the names of corresponding columns in the dataframe

        Returns
        -------
        None

        Raises
        ------
        TypeError
            If any of the input arguments are not of the expected type.
        ValueError
            If any of the input arguments do not have the expected value.
        """   
        super().__init__(data=data,mandatory_properties=mandatory_properties,optional_properties =optional_properties)        
        self._initLog() 


        
    def copy(self):
        """
        Creates a copy of the current RawEventData object.

        Returns
        -------
        copied_object : RawEventData
            A copy of the current RawEventData object.

        """
        # Call the copy method of the base class
        copied_base_object = super().copy()
        # Cast the copied object to RawEventData
        copied_object = RawEventData(copied_base_object.data,
                                      copied_base_object.mandatory_properties,
                                      copied_base_object.optional_properties)      
        return copied_object

    def _filter_dataframe(self, original_dataframe, column_dict):
        """
        Filters a dataframe based on a column names provided in the input dictionary.

        Parameters
        ----------
        original_dataframe : pandas.DataFrame
            The original dataframe to be filtered.
        column_dict : dict
            A dictionary mapping column names to property names.

        Returns
        -------
        subset_dataframe : pandas.DataFrame
            A filtered version of the original dataframe.

    """
        relevant_columns = [col for col in original_dataframe.columns if col in column_dict.values()]
        subset_dataframe = original_dataframe[relevant_columns]
        return subset_dataframe
    
    # Filter the event log according to the specified lifecycle transition types
    def filterLifecycleEvents(self, lifecycleTypes)->'RawEventData':
        """
        Filter the data according to the provided list of desired event lifecycle types, and return a new data object containing only the chosen lifecycle event types

        Parameters
        ----------
        lifecycleTypes : List of chosen lifecycleTypes events (such as 'complete')
        type lifecycleTypes: List

        Returns
        -------
        a new dataobject containing only the chosen event types
        rtype: RawEventData

        """                        
        if lifecycleTypes is None:
            lifecycleTypes = [LifecycleTypes.COMPLETE.value]
        filtered_df = self.data[self.data[self.mandatory_properties[Constants.TYPE_KEY]].str.lower().isin(lifecycleType.lower() for lifecycleType in lifecycleTypes)]   
        return RawEventData(filtered_df, self.mandatory_properties, self.optional_properties)
    
    def getVariants(self)->dict:
        """
        Get all variants in the event log, along with number of traces for each variant

        Returns
        -------
        A dictionary where each key represents a variant name, which is a comma-separated list of all activities in the variant in the order of occurence,and the corresponding value is the number of traces for that variant.

        """
        self._permutations = self._getVariants()        
        count_dict = {key: len(value) for key, value in self._permutations.items()}        
        return count_dict
    
    def getVariantsKeys(self)->dict:
        """
        Get the names of all variants in the event log

        Returns
        -------
        A dictionary where each key represents a variant name, which is a comma-separated list of all activities in the variant in the order of occurence, and the corresponding value is a list of all traces case-ids in this variant.

        """
        self._permutations = self._getVariants()        
        return self._permutations
    
    def _getVariants(self)->dict:
        """
        Build a dictionary representing all process variants in the event log using process mining techniques

        Returns
        -------
        A dictionary where each key represents a variant name joined from all activities within the variant in the order of occurence, and the corresponding value is a list of all trace case-ids in this variant.

        """
        data_map = {}

        if Constants.TYPE_KEY in self.getMandatoryProperties():               
            event_log= self.filterLifecycleEvents([LifecycleTypes.COMPLETE.value])
        else:
            event_log = self
                                   
        formatted_log = event_log.getLog()
        variants = variants_filter.get_variants(formatted_log)

        trace_variant_index=list(variants.keys())
        data_map={}
        for idx, key in enumerate(trace_variant_index):
            case_variant_link=[]
            for trace in variants[key]:
                    case_variant_link.append(trace.attributes[Constants.ACTIVITY_KEY])
            #converting the tuple to string value            
            string_key=",".join(key)
            data_map[string_key]=case_variant_link
        return data_map    
 
        
            
    def getMandatoryPropertiesData(self):
        """
        Return the mandatory properties columns: caseID, activity and event lifecycle columns and timestamp columns
        
        Parameters
        ----------
        None

        Returns
        -------        
        :return: mandatory properties data
        :rtype: _type_
        """        
        return self._filter_dataframe(self.data, self.mandatory_properties)
    
    def getOptionalPropertiesData(self):
        """
        Return the mandatory properties columns: caseID, activity and event lifecycle columns and timestamp columns
        
        Parameters
        ----------
        None

        Returns
        -------        
        :return: mandatory properties data
        :rtype: _type_
        """        
        return self._filter_dataframe(self.data, self.optional_properties)
    
    def _initLog(self):
        """
        Validates that the data in the object is proper event log - has all the mandatory columns, the values of which are of the proper type

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        df = self.getData().copy()
        if type(df) not in [pd.DataFrame]: raise Exception("The method can be applied only to a dataframe!")

        case_id = self.mandatory_properties[Constants.CASE_ID_KEY]
        activity_key = self.mandatory_properties[Constants.ACTIVITY_KEY]
        timestamp_key = self.mandatory_properties[Constants.TIMESTAMP_KEY]
    
        if case_id not in df.columns:
            raise Exception(case_id + " column (case ID) is not in the dataframe!")
        if activity_key not in df.columns:
            raise Exception(activity_key + " column (activity) is not in the dataframe!")
        if timestamp_key not in df.columns:
            raise Exception(timestamp_key + " column (timestamp) is not in the dataframe!")
   
        # make sure the case ID column is of string type
        df[case_id] = df[case_id].astype("string")
        # make sure the activity column is of string type
        df[activity_key] = df[activity_key].astype("string")

        self.data = df
   
    def getLog(self): 
        """
        Return the pm4py event log representing this dataframe object

        Parameters
        ----------
        None

        Returns
        -------
        log : pm4py event log
            The pm4py event log representing this dataframe object

        """     
        df = self.getData().copy()
        if type(df) not in [pd.DataFrame]: raise Exception("The method can be applied only to a dataframe!")

        case_id = self.mandatory_properties[Constants.CASE_ID_KEY]
        activity_key = self.mandatory_properties[Constants.ACTIVITY_KEY]
        timestamp_key = self.mandatory_properties[Constants.TIMESTAMP_KEY]
    
        if case_id not in df.columns:
            raise Exception(case_id + " column (case ID) is not in the dataframe!")
        if activity_key not in df.columns:
            raise Exception(activity_key + " column (activity) is not in the dataframe!")
        if timestamp_key not in df.columns:
            raise Exception(timestamp_key + " column (timestamp) is not in the dataframe!")
   
   
        # make sure the case ID column is of string type
        df[case_id] = df[case_id].astype("string")
        # make sure the activity column is of string type
        df[activity_key] = df[activity_key].astype("string")
   
        name_mapping = {
            case_id: Constants.CASE_ID_KEY,
            activity_key: Constants.ACTIVITY_KEY,
             timestamp_key: Constants.TIMESTAMP_KEY       
        }

        # Rename the columns using the dictionary
        df.rename(columns=name_mapping, inplace=True)       
        return log_converter.apply(df)
    
    
    
  
    def getVariant(self, variant_key) -> 'RawEventData':   
        """
        Get a specific variant data.

        Parameters
        ----------
        variant_key : str
            The name of the variant to retrieve.

        Returns
        -------
        variant : RawEventData
            The RawEventData object representing the specified variant subset of traces.

        """            
        if self._permutations is None:
            self._permutations = self._getVariants()

        chosen_ids = self._permutations[variant_key]        
        id_columns = self.data[self.mandatory_properties[Constants.CASE_ID_KEY]]        
        selected_df = self.data[id_columns.isin(chosen_ids)]                         
        original_df = selected_df.copy()       
        
        return RawEventData(original_df, self.mandatory_properties, self.optional_properties)        
    
    
    
    def transposeToTabular(self) -> TabularEventData:
        """
        Transposes the provided event log data object to a new data object where each trace is represented by a single row (instead of a row per activity),
         where the columns are activity names and the values in the columns are the timestamps of those activities end time.

        Returns
        -------
        data object in tabular format : TabularEventData
            A new data object in tabular format.

        """      
        
        dataFrame = self.getData()
        mandatory_propreties = self.getMandatoryProperties()              
        optional_properties_list =  list(self.getOptionalProperties().values())

        if Constants.TYPE_KEY in mandatory_propreties:
            columns_list = [mandatory_propreties[Constants.CASE_ID_KEY],mandatory_propreties[Constants.ACTIVITY_KEY],mandatory_propreties[Constants.TYPE_KEY],mandatory_propreties[Constants.TIMESTAMP_KEY]]
        else:
            columns_list = [mandatory_propreties[Constants.CASE_ID_KEY],mandatory_propreties[Constants.ACTIVITY_KEY],mandatory_propreties[Constants.TIMESTAMP_KEY]]
        columns_list+=optional_properties_list  
                

        new_df= self._flat_sources_updated(dataFrame,optional_properties_list,mandatory_propreties)

        id_column= mandatory_propreties[Constants.CASE_ID_KEY]
        start_time_column = mandatory_propreties[Constants.STARTTIME_COLUMN]       
        
        unique_start_times = dataFrame[[id_column, start_time_column]].drop_duplicates(subset=id_column, keep='first')              
        new_df = pd.merge(new_df,unique_start_times, on = id_column, how='left')
        new_df = new_df[[id_column,  start_time_column] + [col for col in new_df.columns if col not in [id_column, start_time_column]]]

        mandatory_columns_names,optional_column_names = self._extract_properties(new_df,self.getMandatoryProperties())

        new_df = new_df.reset_index(drop=True)   
        
        
        return  TabularEventData(new_df,mandatory_columns_names,optional_column_names)
    
    def transposeFullDataframe(self) -> DataFrame:
        mandatory_properties = self.getMandatoryProperties()
        transposedMandatory = self.transposeToTabular().getData()
        transposedOptional = self._transposeToTabularOptionalProperties()
        merged_df = pd.merge(transposedMandatory, transposedOptional, on=mandatory_properties[Constants.CASE_ID_KEY], how='inner')
        return merged_df
    
    def _transposeToTabularOptionalProperties(self) -> DataFrame:
        mandatory_properties = self.getMandatoryProperties()
        optional_properties = self.getOptionalProperties()
        
        new_data = {}
        df = self.getData()
        for _, row in df.iterrows():
            case_id = row[mandatory_properties[Constants.CASE_ID_KEY]]
            activity = row[mandatory_properties[Constants.ACTIVITY_KEY]]
    
            if case_id not in new_data:
                new_data[case_id] = {}
    
            for column in df.columns:
                if column != mandatory_properties[Constants.CASE_ID_KEY] and column != mandatory_properties[Constants.ACTIVITY_KEY] and column in optional_properties:
                    new_column_name = f"{activity}_{column}"
                    new_data[case_id][new_column_name] = row[column]
        
        new_df = pd.DataFrame.from_dict(new_data, orient='index').reset_index().rename(columns={'index': mandatory_properties[Constants.CASE_ID_KEY]})
        print(new_df)
        return new_df

    def _flat_sources_updated(self,x,columns,mandatory_propreties) -> DataFrame:
        """
        Pivots the dataframe and returns a new dataframe with the pivoted data.

        Parameters
        ----------
        x : DataFrame
            The input dataframe to be pivoted.
        columns : list
            A list of column names to be used as the new column names.
        mandatory_propreties : dict
            A dictionary containing the mandatory properties of the data object.

        Returns
        -------
        DataFrame
            A new dataframe with the pivoted data.

        """
        # Pivot the dataframe
        new_x = pd.DataFrame() 
        #get variants in this dataframe and perform this for each variant        
        Id_column_name = mandatory_propreties[Constants.CASE_ID_KEY]
        Activity_column_name = mandatory_propreties[Constants.ACTIVITY_KEY]
        Timestamp_column_name = mandatory_propreties[Constants.TIMESTAMP_KEY]
        #taking care of loops - taking the last loop
        #TODO: add function with config how to treat loops. Possible option: last loop, remove variants with loops altogether, "open" loops. Right now takes max timestamp value - last occurence
        pivoted_df = x.pivot_table(index=Id_column_name, columns=Activity_column_name, values=Timestamp_column_name, aggfunc='max', fill_value=None)


        # Reset index to make 'caseId' a column again
        pivoted_df.reset_index(inplace=True)

        # Rename the columns
        pivoted_df.columns.name = None  # Remove the name of the columns index
        pivoted_df.columns = [Id_column_name] + pivoted_df.columns[1:].tolist()
        return pivoted_df


       
        
        
                        
        



        